import random
import pytest
from typing import List, Optional
from harness.src.network import Network
from rraft import (
    NO_LIMIT,
    ConfChange,
    ConfChangeTransition,
    ConfChangeType,
    ConfChangeV2,
    ConfState,
    Config,
    ConfigRef,
    Entry,
    EntryRef,
    EntryType,
    GetEntriesContext,
    HardStateRef,
    LoggerRef,
    MemStorage,
    Message,
    MessageType,
    InMemoryRawNode,
    InMemoryRawNodeRef,
    ProposalDroppedError,
    ReadState,
    ReadyRef,
    Snapshot,
    SnapshotRef,
    SoftStateRef,
    StateRole,
    default_logger,
    new_conf_change_single,
    is_local_msg,
)
from harness.utils import (
    conf_change_v2,
    conf_change,
    conf_state,
    conf_state_v2,
    empty_entry,
    new_entry,
    new_message_with_entries,
    new_message,
    new_snapshot,
    new_storage,
    new_test_config,
    new_test_raft_with_config,
    new_test_raft,
    SOME_DATA,
    hard_state,
    remove_node,
    soft_state,
)


MAX_UINT64 = (1 << 64) - 1


def must_cmp_ready(
    r: ReadyRef,
    ss: Optional[SoftStateRef],
    hs: Optional[HardStateRef],
    entries: List[EntryRef],
    committed_entries: List[EntryRef],
    snapshot: Optional[SnapshotRef],
    msg_is_empty: bool,
    persisted_msg_is_empty: bool,
    must_sync: bool,
):
    assert r.ss() == ss
    assert r.hs() == hs
    assert r.entries() == entries
    assert r.committed_entries() == committed_entries
    assert r.must_sync() == must_sync
    assert r.snapshot() == snapshot or r.snapshot() == Snapshot.default()
    assert len(r.messages()) == 0 if msg_is_empty else len(r.messages()) != 0
    assert (
        len(r.persisted_messages()) == 0
        if persisted_msg_is_empty
        else len(r.persisted_messages()) != 0
    )


def new_raw_node(
    id: int,
    peers: List[int],
    election_tick: int,
    heartbeat_tick: int,
    storage: MemStorage,
    logger: LoggerRef,
) -> InMemoryRawNode:
    config = new_test_config(id, election_tick, heartbeat_tick)
    return new_raw_node_with_config(peers, config, storage, logger)


def new_raw_node_with_config(
    peers: List[int],
    config: ConfigRef,
    storage: MemStorage,
    logger: LoggerRef,
) -> InMemoryRawNode:
    initial_state = storage.initial_state()
    assert not (
        initial_state.initialized() and not peers
    ), f"new_raw_node with empty peers on initialized store"

    if peers and not initial_state.initialized():
        storage.wl().apply_snapshot(new_snapshot(1, 1, peers))

    return InMemoryRawNode(config, storage, logger)


def get_msg_types() -> List[MessageType]:
    return [
        MessageType.MsgHup,
        MessageType.MsgBeat,
        MessageType.MsgPropose,
        MessageType.MsgAppend,
        MessageType.MsgAppendResponse,
        MessageType.MsgRequestVote,
        MessageType.MsgRequestVoteResponse,
        MessageType.MsgSnapshot,
        MessageType.MsgHeartbeat,
        MessageType.MsgHeartbeatResponse,
        MessageType.MsgUnreachable,
        MessageType.MsgSnapStatus,
        MessageType.MsgCheckQuorum,
        MessageType.MsgTransferLeader,
        MessageType.MsgTimeoutNow,
        MessageType.MsgReadIndex,
        MessageType.MsgReadIndexResp,
        MessageType.MsgRequestPreVote,
        MessageType.MsgRequestPreVoteResponse,
    ]


# Ensures that RawNode::step ignore local message.
def test_raw_node_step():
    l = default_logger()
    for msg_t in get_msg_types():
        s = new_storage()
        s.wl().set_hardstate(hard_state(1, 1, 0))
        # Append an empty entry to make sure the non-local messages (like
        # vote requests) are ignored and don't trigger assertions.
        s.wl().append([new_entry(1, 1, None)])
        s.wl().apply_snapshot(new_snapshot(1, 1, [1]))

        storage = new_storage()
        raw_node = new_raw_node(1, [1], 10, 1, storage, l)
        # LocalMsg should be ignored.
        try:
            raw_node.step(new_message(0, 0, msg_t, 0))
        except Exception as e:
            if is_local_msg(msg_t):
                assert str(e) == "raft: cannot step raft local message"


# Ensures that MsgReadIndex to old leader gets forwarded to the new leader and
# 'send' method does not attach its term.
def test_raw_node_read_index_to_old_leader():
    l = default_logger()
    s1, s2, s3 = new_storage(), new_storage(), new_storage()
    r1 = new_test_raft(1, [1, 2, 3], 10, 1, s1, l)
    r2 = new_test_raft(2, [1, 2, 3], 10, 1, s2, l)
    r3 = new_test_raft(3, [1, 2, 3], 10, 1, s3, l)

    nt = Network.new([r1, r2, r3], l)

    # elect r1 as leader
    nt.send([new_message(1, 1, MessageType.MsgHup, 0)])

    test_entries = Entry.default()
    test_entries.set_data(b"testdata")

    # send readindex request to r2(follower)
    _ = nt.peers.get(2).step(
        new_message_with_entries(
            2,
            2,
            MessageType.MsgReadIndex,
            [test_entries.clone()],
        )
    )

    # verify r2(follower) forwards this message to r1(leader) with term not set
    assert len(nt.peers[2].raft.get_msgs()) == 1
    read_index_msg1 = new_message_with_entries(
        2, 1, MessageType.MsgReadIndex, [test_entries.clone()]
    )
    assert read_index_msg1 == nt.peers[2].raft.get_msgs()[0]

    # send readindex request to r3(follower)
    _ = nt.peers.get(3).step(
        new_message_with_entries(
            3,
            3,
            MessageType.MsgReadIndex,
            [test_entries.clone()],
        )
    )

    # verify r3(follower) forwards this message to r1(leader) with term not set as well.
    assert len(nt.peers[3].raft.get_msgs()) == 1

    read_index_msg2 = new_message_with_entries(
        3, 1, MessageType.MsgReadIndex, [test_entries.clone()]
    )
    assert nt.peers[3].raft.get_msgs()[0] == read_index_msg2

    # now elect r3 as leader
    nt.send([new_message(3, 3, MessageType.MsgHup, 0)])

    # let r1 steps the two messages previously we got from r2, r3
    nt.peers.get(1).step(read_index_msg1)
    nt.peers.get(1).step(read_index_msg2)

    # verify r1(follower) forwards these messages again to r3(new leader)
    assert len(nt.peers[1].raft.get_msgs()) == 2
    assert nt.peers[1].raft.get_msgs()[0] == new_message_with_entries(
        2, 3, MessageType.MsgReadIndex, [test_entries.clone()]
    )

    assert nt.peers[1].raft.get_msgs()[1] == new_message_with_entries(
        3, 3, MessageType.MsgReadIndex, [test_entries]
    )


# Tests the configuration change mechanism. Each test case sends a configuration
# change which is either simple or joint, verifies that it applies and that the
# resulting ConfState matches expectations, and for joint configurations makes
# sure that they are exited successfully.
def test_raw_node_propose_and_conf_change():
    l = default_logger()

    class Test:
        def __init__(
            self,
            cc: ConfChange | ConfChangeV2,
            exp: ConfState,
            exp2: Optional[ConfState],
        ) -> None:
            self.cc = cc
            self.exp = exp
            self.exp2 = exp2

    test_cases = [
        # V1 config change.b
        Test(
            conf_change(ConfChangeType.AddNode, 2),
            conf_state([1, 2], []),
            None,
        ),
    ]

    # Proposing the same as a V2 change works just the same, without entering
    # a joint config.
    single = new_conf_change_single(2, ConfChangeType.AddNode)
    test_cases.append(
        Test(
            conf_change_v2([single]),
            conf_state([1, 2], []),
            None,
        )
    )

    # Ditto if we add it as a learner instead.
    single = new_conf_change_single(2, ConfChangeType.AddLearnerNode)
    test_cases.append(
        Test(
            conf_change_v2([single]),
            conf_state([1], [2]),
            None,
        )
    )

    # We can ask explicitly for joint consensus if we want it.
    single = new_conf_change_single(2, ConfChangeType.AddLearnerNode)
    cc = conf_change_v2([single])
    cc.set_transition(ConfChangeTransition.Explicit)
    cs = conf_state_v2([1], [2], [1], [], False)
    test_cases.append(Test(cc, cs, conf_state([1], [2])))

    # Ditto, but with implicit transition (the harness checks this).
    single = new_conf_change_single(2, ConfChangeType.AddLearnerNode)
    cc = conf_change_v2([single])
    cc.set_transition(ConfChangeTransition.Implicit)
    cs = conf_state_v2([1], [2], [1], [], True)
    test_cases.append(Test(cc, cs, conf_state([1], [2])))

    # Add a new node and demote n1. This exercises the interesting case in
    # which we really need joint config changes and also need LearnersNext.
    cc = conf_change_v2(
        [
            new_conf_change_single(2, ConfChangeType.AddNode),
            new_conf_change_single(1, ConfChangeType.AddLearnerNode),
            new_conf_change_single(3, ConfChangeType.AddLearnerNode),
        ]
    )
    cs = conf_state_v2([2], [3], [1], [1], True)
    test_cases.append(Test(cc, cs, conf_state([2], [1, 3])))

    # Ditto explicit.
    cc = conf_change_v2(
        [
            new_conf_change_single(2, ConfChangeType.AddNode),
            new_conf_change_single(1, ConfChangeType.AddLearnerNode),
            new_conf_change_single(3, ConfChangeType.AddLearnerNode),
        ]
    )
    cc.set_transition(ConfChangeTransition.Explicit)
    cs = conf_state_v2([2], [3], [1], [1], False)
    test_cases.append(Test(cc, cs, conf_state([2], [1, 3])))

    # Ditto implicit.
    cc = conf_change_v2(
        [
            new_conf_change_single(2, ConfChangeType.AddNode),
            new_conf_change_single(1, ConfChangeType.AddLearnerNode),
            new_conf_change_single(3, ConfChangeType.AddLearnerNode),
        ]
    )
    cc.set_transition(ConfChangeTransition.Implicit)
    cs = conf_state_v2([2], [3], [1], [1], True)
    test_cases.append(Test(cc, cs, conf_state([2], [1, 3])))

    for v in test_cases:
        cc, exp, exp2 = v.cc, v.exp, v.exp2
        s = new_storage()

        raw_node = new_raw_node(1, [1], 10, 1, s.clone(), l)
        raw_node.campaign()
        proposed = False
        ccdata = []
        # Propose the ConfChange, wait until it applies, save the resulting ConfState.
        cs: Optional[ConfState] = None

        while not cs:
            rd = raw_node.ready()
            s.wl().append(rd.entries())

            def handle_committed_entries(
                rn: InMemoryRawNodeRef, committed_entries: List[Entry]
            ):
                for e in committed_entries:
                    nonlocal cs

                    if e.get_entry_type() == EntryType.EntryConfChange:
                        cc = ConfChange.default()
                        cc.merge_from_bytes(e.get_data())
                        cs = rn.apply_conf_change(cc)
                    elif e.get_entry_type() == EntryType.EntryConfChangeV2:
                        cc = ConfChangeV2.default()
                        cc.merge_from_bytes(e.get_data())
                        cs = rn.apply_conf_change_v2(cc)

            handle_committed_entries(raw_node.make_ref(), rd.take_committed_entries())

            is_leader = False

            if ss := rd.ss():
                is_leader = ss.get_leader_id() == raw_node.get_raft().get_id()

            light_rd = raw_node.advance(rd.make_ref())
            handle_committed_entries(
                raw_node.make_ref(), light_rd.take_committed_entries()
            )
            raw_node.advance_apply()

            # Once we are the leader, propose a command and a ConfChange.
            if not proposed and is_leader:
                raw_node.propose([], b"somedata")

                if v1 := cc.as_v1():
                    ccdata = v1.encode()
                    raw_node.propose_conf_change([], v1.clone())
                else:
                    v2 = cc.as_v2()
                    ccdata = v2.encode()
                    raw_node.propose_conf_change_v2([], v2)

                proposed = True

        # Check that the last index is exactly the conf change we put in,
        # down to the bits. Note that this comes from the Storage, which
        # will not reflect any unstable entries that we'll only be presented
        # with in the next Ready.
        last_index = s.last_index()
        ctx = GetEntriesContext.empty(False)
        entries = s.entries(last_index - 1, last_index + 1, ctx.make_ref(), NO_LIMIT)
        assert len(entries) == 2
        assert entries[0].get_data() == b"somedata"

        if cc.as_v1():
            assert entries[1].get_entry_type() == EntryType.EntryConfChange
        else:
            assert entries[1].get_entry_type() == EntryType.EntryConfChangeV2

        assert ccdata == entries[1].get_data()
        assert cs == exp

        conf_index = last_index
        v2 = cc.as_v2()
        if v2.enter_joint():
            # If this is an auto-leaving joint conf change, it will have
            # appended the entry that auto-leaves, so add one to the last
            # index that forms the basis of our expectations on
            # pendingConfIndex. (Recall that lastIndex was taken from stable
            # storage, but this auto-leaving entry isn't on stable storage
            # yet).
            conf_index += 1

        assert conf_index == raw_node.get_raft().get_pending_conf_index()

        # Move the RawNode along. If the ConfChange was simple, nothing else
        # should happen. Otherwise, we're in a joint state, which is either
        # left automatically or not. If not, we add the proposal that leaves
        # it manually.
        rd = raw_node.ready()
        context = []
        if not exp.get_auto_leave():
            assert not rd.entries()
            if not exp2:
                continue

            context = list(b"manual")
            cc = conf_change_v2([])
            cc.set_context(context)
            raw_node.propose_conf_change_v2([], cc)
            rd = raw_node.ready()

        # Check that the right ConfChange comes out.
        assert len(rd.entries()) == 1
        assert rd.entries()[0].get_entry_type() == EntryType.EntryConfChangeV2
        leave_cc = ConfChangeV2.default()
        leave_cc.merge_from_bytes(rd.entries()[0].get_data())

        assert context == list(leave_cc.get_context()), f"{cc.as_v2()}"
        # Lie and pretend the ConfChange applied. It won't do so because now
        # we require the joint quorum and we're only running one node.
        cs = raw_node.apply_conf_change_v2(leave_cc)
        assert cs == exp2


# Tests the configuration change auto leave even leader lost leadership.
def test_raw_node_joint_auto_leave():
    l = default_logger()

    single = new_conf_change_single(2, ConfChangeType.AddLearnerNode)
    test_cc = conf_change_v2([single])
    test_cc.set_transition(ConfChangeTransition.Implicit)
    exp_cs = conf_state_v2([1], [2], [1], [], True)
    exp_cs2 = conf_state([1], [2])

    s = new_storage()
    raw_node = new_raw_node(1, [1], 10, 1, s.clone(), l)
    raw_node.campaign()
    proposed = False
    ccdata = test_cc.encode()
    # Propose the ConfChange, wait until it applies, save the resulting ConfState.
    cs = None
    while not cs:
        rd = raw_node.ready()
        s.wl().append(rd.entries())

        def handle_committed_entries(
            rn: InMemoryRawNodeRef, committed_entries: List[Entry]
        ):
            for e in committed_entries:
                nonlocal cs
                if e.get_entry_type() == EntryType.EntryConfChangeV2:
                    cc = ConfChangeV2.default()
                    cc.merge_from_bytes(e.get_data())

                    # Force it step down.
                    msg = new_message(1, 1, MessageType.MsgHeartbeatResponse, 0)
                    msg.set_term(rn.get_raft().get_term() + 1)
                    rn.step(msg)

                    cs = rn.apply_conf_change_v2(cc)

        handle_committed_entries(raw_node.make_ref(), rd.take_committed_entries())
        is_leader = False

        if ss := rd.ss():
            is_leader = ss.get_leader_id() == raw_node.get_raft().get_id()

        light_rd = raw_node.advance(rd.make_ref())
        handle_committed_entries(raw_node.make_ref(), light_rd.take_committed_entries())
        raw_node.advance_apply()

        # Once we are the leader, propose a command and a ConfChange.
        if not proposed and is_leader:
            raw_node.propose([], b"somedata")
            raw_node.propose_conf_change_v2([], test_cc.clone())

            proposed = True

    # Check that the last index is exactly the conf change we put in,
    # down to the bits. Note that this comes from the Storage, which
    # will not reflect any unstable entries that we'll only be presented
    # with in the next Ready.
    last_index = s.last_index()
    ctx = GetEntriesContext.empty(False)
    entries = s.entries(last_index - 1, last_index + 1, ctx.make_ref(), NO_LIMIT)
    assert len(entries) == 2
    assert entries[0].get_data() == b"somedata"
    assert entries[1].get_entry_type() == EntryType.EntryConfChangeV2
    assert ccdata == entries[1].get_data()
    assert exp_cs == cs
    assert raw_node.get_raft().get_pending_conf_index() == 0

    # Move the RawNode along. It should not leave joint because it's follower.
    rd = raw_node.ready()
    assert not rd.entries()
    _ = raw_node.advance(rd.make_ref())

    # Make it leader again. It should leave joint automatically after moving apply index.
    raw_node.campaign()
    rd = raw_node.ready()
    s.wl().append(rd.entries())
    _ = raw_node.advance(rd.make_ref())

    rd = raw_node.ready()
    s.wl().append(rd.entries())

    # Check that the right ConfChange comes out.
    assert len(rd.entries()) == 1
    assert rd.entries()[0].get_entry_type() == EntryType.EntryConfChangeV2
    leave_cc = ConfChangeV2.default()
    leave_cc.merge_from_bytes(rd.entries()[0].get_data())

    assert not leave_cc.get_context()

    # Lie and pretend the ConfChange applied. It won't do so because now
    # we require the joint quorum and we're only running one node.
    cs = raw_node.apply_conf_change_v2(leave_cc)
    assert cs == exp_cs2


# Ensures that two proposes to add the same node should not affect the later propose
# to add new node.
def test_raw_node_propose_add_duplicate_node():
    l = default_logger()
    s = new_storage()
    raw_node = new_raw_node(1, [1], 10, 1, s.clone(), l)
    raw_node.campaign()

    while True:
        rd = raw_node.ready()
        s.wl().append(rd.entries())

        if ss := rd.ss():
            if ss.get_leader_id() == raw_node.get_raft().get_id():
                raw_node.advance(rd.make_ref())
                break
        raw_node.advance(rd.make_ref())

    def propose_conf_change_and_apply(cc: ConfChange):
        raw_node.propose_conf_change([], cc)
        rd = raw_node.ready()
        s.wl().append(rd.entries())

        def handle_committed_entries(
            rn: InMemoryRawNodeRef, committed_entries: List[Entry]
        ):
            for e in committed_entries:
                if e.get_entry_type() == EntryType.EntryConfChange:
                    conf_change = ConfChange.default()
                    conf_change.merge_from_bytes(e.get_data())
                    rn.apply_conf_change(conf_change)

        handle_committed_entries(raw_node, rd.take_committed_entries())

        light_rd = raw_node.advance(rd.make_ref())
        handle_committed_entries(raw_node, light_rd.take_committed_entries())
        raw_node.advance_apply()

    cc1 = conf_change(ConfChangeType.AddNode, 1)
    ccdata1 = cc1.encode()
    propose_conf_change_and_apply(cc1.clone())

    # try to add the same node again
    propose_conf_change_and_apply(cc1)

    # the new node join should be ok
    cc2 = conf_change(ConfChangeType.AddNode, 2)
    ccdata2 = cc2.encode()
    propose_conf_change_and_apply(cc2)

    last_index = s.last_index()

    # the last three entries should be: ConfChange cc1, cc1, cc2
    ctx = GetEntriesContext.empty(False)
    entries = s.entries(last_index - 2, last_index + 1, ctx.make_ref(), None)
    assert len(entries) == 3
    assert entries[0].get_data() == ccdata1
    assert entries[2].get_data() == ccdata2


def test_raw_node_propose_add_learner_node():
    l = default_logger()
    s = new_storage()
    raw_node = new_raw_node(1, [1], 10, 1, s.clone(), l)
    rd = raw_node.ready()
    must_cmp_ready(rd.make_ref(), None, None, [], [], None, True, True, False)
    _ = raw_node.advance(rd.make_ref())

    raw_node.campaign()

    while True:
        rd = raw_node.ready()
        if ss := rd.ss():
            if ss.get_leader_id() == raw_node.get_raft().get_id():
                raw_node.advance(rd.make_ref())
                break

    # propose add learner node and check apply state
    cc = conf_change(ConfChangeType.AddLearnerNode, 2)
    raw_node.propose_conf_change([], cc)

    rd = raw_node.ready()
    s.wl().append(rd.entries())

    light_rd = raw_node.advance(rd.make_ref())

    assert (
        len(light_rd.committed_entries()) == 1
    ), f"should committed the conf change entry"

    e = light_rd.committed_entries()[0]
    assert e.get_entry_type() == EntryType.EntryConfChange
    conf_change_ = ConfChange.default()
    conf_change_.merge_from_bytes(e.get_data())
    conf_state = raw_node.apply_conf_change(conf_change_)
    assert conf_state.get_voters() == [1]
    assert conf_state.get_learners() == [2]


# Ensures that RawNode.read_index sends the MsgReadIndex message to the underlying
# raft. It also ensures that ReadState can be read out.
def test_raw_node_read_index():
    l = default_logger()
    wrequest_ctx = list(b"somedata")
    wrs = [ReadState.default()]
    wrs[0].set_index(2)
    wrs[0].set_request_ctx(wrequest_ctx)

    s = new_storage()
    raw_node = new_raw_node(1, [1], 10, 1, s.clone(), l)
    raw_node.campaign()
    while True:
        rd = raw_node.ready()
        s.wl().append(rd.entries())

        if ss := rd.ss():
            if ss.get_leader_id() == raw_node.get_raft().get_id():
                raw_node.advance(rd.make_ref())

                # Once we are the leader, issue a read index request
                raw_node.read_index(wrequest_ctx)
                break

        raw_node.advance(rd.make_ref())

    # ensure the read_states can be read out
    assert raw_node.get_raft().get_read_states()
    assert raw_node.has_ready()
    rd = raw_node.ready()
    assert rd.read_states() == wrs
    s.wl().append(rd.entries())
    raw_node.advance(rd.make_ref())

    # ensure raft.read_states is reset after advance
    assert not raw_node.has_ready()
    assert not raw_node.get_raft().get_read_states()


# Ensures that a node can be started correctly. Note that RawNode requires the
# application to bootstrap the state, i.e. it does not accept peers and will not
# create faux configuration change entries.
def test_raw_node_start():
    l = default_logger()
    store = new_storage()
    raw_node = new_raw_node(1, [1], 10, 1, store.clone(), l)

    rd = raw_node.ready()
    must_cmp_ready(rd, None, None, [], [], None, True, True, False)
    _ = raw_node.advance(rd.make_ref())

    raw_node.campaign()
    rd = raw_node.ready()
    ss = soft_state(1, StateRole.Leader)

    must_cmp_ready(
        rd.make_ref(),
        ss.make_ref(),
        hard_state(2, 1, 1),
        [new_entry(2, 2, None)],
        [],
        None,
        True,
        True,
        True,
    )
    store.wl().append(rd.entries())

    light_rd = raw_node.advance(rd.make_ref())
    assert light_rd.commit_index() == 2
    assert light_rd.committed_entries() == [new_entry(2, 2, None)]
    assert not raw_node.has_ready()

    raw_node.propose([], list(b"somedata"))
    rd = raw_node.ready()
    must_cmp_ready(
        rd.make_ref(),
        None,
        None,
        [new_entry(2, 3, SOME_DATA)],
        [],
        None,
        True,
        True,
        True,
    )
    store.wl().append(rd.entries())
    light_rd = raw_node.advance(rd.make_ref())
    assert light_rd.commit_index() == 3
    assert light_rd.committed_entries() == [new_entry(2, 3, SOME_DATA)]

    assert not raw_node.has_ready()


def test_raw_node_restart():
    l = default_logger()
    entries = [empty_entry(1, 1), new_entry(1, 2, "foo")]

    store = new_storage()
    store.wl().set_hardstate(hard_state(1, 1, 0))
    store.wl().append(entries)
    raw_node = new_raw_node(1, [], 10, 1, store, l)

    rd = raw_node.ready()
    must_cmp_ready(
        rd,
        None,
        None,
        [],
        entries[:1],
        None,
        True,
        True,
        False,
    )
    raw_node.advance(rd.make_ref())
    assert not raw_node.has_ready()


def test_raw_node_restart_from_snapshot():
    l = default_logger()
    snap = new_snapshot(2, 1, [1, 2])
    entries = [new_entry(1, 3, "foo")]

    store = new_storage()
    store.wl().apply_snapshot(snap)
    store.wl().append(entries)
    store.wl().set_hardstate(hard_state(1, 3, 0))
    raw_node = InMemoryRawNode(new_test_config(1, 10, 1), store, l)

    rd = raw_node.ready()
    must_cmp_ready(rd.make_ref(), None, None, [], entries, None, True, True, False)
    raw_node.advance(rd.make_ref())
    assert not raw_node.has_ready()


# test_skip_bcast_commit ensures that empty commit message is not sent out
# when skip_bcast_commit is true.
def test_skip_bcast_commit():
    l = default_logger()
    config = new_test_config(1, 10, 1)
    config.set_skip_bcast_commit(True)
    s1 = MemStorage.new_with_conf_state(ConfState([1, 2, 3], []))
    r1 = new_test_raft_with_config(config, s1, l)
    s2 = new_storage()
    r2 = new_test_raft(2, [1, 2, 3], 10, 1, s2, l)
    s3 = new_storage()
    r3 = new_test_raft(3, [1, 2, 3], 10, 1, s3, l)
    nt = Network.new([r1, r2, r3], l)

    # elect r1 as leader
    nt.send([new_message(1, 1, MessageType.MsgHup, 0)])

    # Without bcast commit, followers will not update its commit index immediately.
    test_entries = Entry.default()
    test_entries.set_data(b"testdata")
    msg = new_message_with_entries(1, 1, MessageType.MsgPropose, [test_entries])
    nt.send([msg.clone()])

    assert nt.peers[1].raft_log.get_committed() == 2
    assert nt.peers[2].raft_log.get_committed() == 1
    assert nt.peers[3].raft_log.get_committed() == 1

    # After bcast heartbeat, followers will be informed the actual commit index.
    for _ in range(0, nt.peers[1].raft.randomized_election_timeout()):
        nt.peers.get(1).raft.tick()

    nt.send([new_message(1, 1, MessageType.MsgHup, 0)])
    assert nt.peers[2].raft_log.get_committed() == 2
    assert nt.peers[3].raft_log.get_committed() == 2

    # The feature should be able to be adjusted at run time.
    nt.peers.get(1).raft.skip_bcast_commit(False)
    nt.send([msg.clone()])
    assert nt.peers[1].raft_log.get_committed() == 3
    assert nt.peers[2].raft_log.get_committed() == 3
    assert nt.peers[3].raft_log.get_committed() == 3

    nt.peers.get(1).raft.skip_bcast_commit(True)

    # Later proposal should commit former proposal.
    nt.send([msg.clone()])
    nt.send([msg])
    assert nt.peers[1].raft_log.get_committed() == 5
    assert nt.peers[2].raft_log.get_committed() == 4
    assert nt.peers[3].raft_log.get_committed() == 4

    # When committing conf change, leader should always bcast commit.
    cc = ConfChange.default()
    cc.set_change_type(ConfChangeType.RemoveNode)
    cc.set_node_id(3)
    data = cc.encode()
    cc_entry = Entry.default()
    cc_entry.set_entry_type(EntryType.EntryConfChange)
    cc_entry.set_data(data)
    nt.send(
        [
            new_message_with_entries(
                1,
                1,
                MessageType.MsgPropose,
                [cc_entry],
            )
        ]
    )

    assert nt.peers[1].raft.should_bcast_commit()
    assert nt.peers[2].raft.should_bcast_commit()
    assert nt.peers[3].raft.should_bcast_commit()

    assert nt.peers[1].raft_log.get_committed() == 6
    assert nt.peers[2].raft_log.get_committed() == 6
    assert nt.peers[3].raft_log.get_committed() == 6


# test_set_priority checks the set_priority function in RawNode.
def test_set_priority():
    l = default_logger()
    s = new_storage()
    raw_node = new_raw_node(1, [1], 10, 1, s, l)
    priorities = [0, 1, 5, 10, 10000]
    for p in priorities:
        raw_node.set_priority(p)
        assert raw_node.get_raft().get_priority() == p


# TestNodeBoundedLogGrowthWithPartition tests a scenario where a leader is
# partitioned from a quorum of nodes. It verifies that the leader's log is
# protected from unbounded growth even as new entries continue to be proposed.
# This protection is provided by the max_uncommitted_size configuration.
def test_bounded_uncommitted_entries_growth_with_partition():
    l = default_logger()
    config = Config(id=1, max_uncommitted_size=12)
    s = new_storage()
    raw_node = new_raw_node_with_config([1], config, s.clone(), l)

    # wait raw_node to be leader
    raw_node.campaign()
    while True:
        rd = raw_node.ready()
        s.wl().set_hardstate(rd.hs().clone())
        s.wl().append(rd.entries())

        if rd.ss():
            raw_node.advance(rd.make_ref())
            break

        raw_node.advance(rd.make_ref())

    # should be accepted
    data = b"hello world!"
    raw_node.propose([], data)

    # should be dropped
    with pytest.raises(ProposalDroppedError):
        raw_node.propose([], data)

    # should be accepted when previous data has been committed
    rd = raw_node.ready()
    s.wl().append(rd.entries())
    raw_node.advance(rd.make_ref())

    data = list(b"hello world!")
    raw_node.propose([], data)


def prepare_async_entries(raw_node: InMemoryRawNode, s: MemStorage):
    raw_node.get_raft().become_candidate()
    raw_node.get_raft().become_leader()

    rd = raw_node.ready()
    s.wl().append(rd.entries())
    raw_node.advance(rd.make_ref())

    data = [1] * 1000
    for _ in range(0, 10):
        raw_node.propose([], data)

    rd = raw_node.ready()
    entries = rd.entries()
    assert len(entries) == 10
    s.wl().append(entries)
    msgs = rd.messages()
    # First append has two entries: the empty entry to confirm the
    # election, and the first proposal (only one proposal gets sent
    # because we're in probe state).
    assert len(msgs) == 1
    assert msgs[0].get_msg_type() == MessageType.MsgAppend
    assert len(msgs[0].get_entries()) == 2
    raw_node.advance_append(rd.make_ref())

    s.wl().trigger_log_unavailable(True)

    # Become replicate state
    append_response = new_message(2, 1, MessageType.MsgAppendResponse, 0)
    append_response.set_term(2)
    append_response.set_index(2)
    raw_node.step(append_response)


# Test entries are handled properly when they are fetched asynchronously
def test_raw_node_with_async_entries():
    l = default_logger()
    cfg = new_test_config(1, 10, 1)
    cfg.set_max_size_per_msg(2048)
    s = new_storage()
    raw_node = new_raw_node_with_config([1, 2], cfg, s, l)

    prepare_async_entries(raw_node, s)

    # No entries are sent because the entries are temporarily unavailable
    rd = raw_node.ready()
    entries = rd.entries()
    s.wl().append(entries)
    msgs = rd.messages()
    assert not msgs
    raw_node.advance_append(rd.make_ref())

    # Entries are sent when the entries are ready which is informed by `on_entries_fetched`.
    s.wl().trigger_log_unavailable(False)
    context = s.wl().take_get_entries_context()
    raw_node.on_entries_fetched(context.make_ref())
    rd = raw_node.ready()
    entries = rd.entries()
    s.wl().append(entries)
    msgs = rd.messages()
    assert len(msgs) == 5
    assert msgs[0].get_msg_type() == MessageType.MsgAppend
    assert len(msgs[0].get_entries()) == 2
    raw_node.advance_append(rd.make_ref())


# Test if async fetch entries works well when there is a remove node conf-change.
def test_raw_node_with_async_entries_to_removed_node():
    l = default_logger()
    cfg = new_test_config(1, 10, 1)
    cfg.set_max_size_per_msg(2048)
    s = new_storage()
    raw_node = new_raw_node_with_config([1, 2], cfg, s.clone(), l)

    prepare_async_entries(raw_node, s)

    raw_node.apply_conf_change_v2(remove_node(2))

    # Entries are not sent due to the node is removed.
    s.wl().trigger_log_unavailable(False)
    context = s.wl().take_get_entries_context()
    raw_node.on_entries_fetched(context.make_ref())
    rd = raw_node.ready()
    assert not rd.entries()
    assert not rd.messages()
    raw_node.advance_append(rd.make_ref())


# Test if async fetch entries works well when there is a leader step-down.
def test_raw_node_with_async_entries_on_follower():
    l = default_logger()
    cfg = new_test_config(1, 10, 1)
    cfg.set_max_size_per_msg(2048)
    s = new_storage()
    raw_node = new_raw_node_with_config([1, 2], cfg, s.clone(), l)

    prepare_async_entries(raw_node, s)

    # Set recent inactive to step down leader
    raw_node.get_raft().prs().get(2).set_recent_active(False)
    msg = Message()
    msg.set_to(1)
    msg.set_msg_type(MessageType.MsgCheckQuorum)
    raw_node.get_raft().step(msg)
    assert raw_node.get_raft().get_state() != StateRole.Leader

    # Entries are not sent due to the leader is changed.
    s.wl().trigger_log_unavailable(False)
    context = s.wl().take_get_entries_context()
    raw_node.on_entries_fetched(context.make_ref())
    rd = raw_node.ready()
    assert not rd.entries()
    assert not rd.messages()
    raw_node.advance_append(rd.make_ref())


def test_raw_node_async_entries_with_leader_change():
    l = default_logger()
    cfg = new_test_config(1, 10, 1)
    cfg.set_max_size_per_msg(2048)
    s = new_storage()
    raw_node = new_raw_node_with_config([1, 2], cfg, s.clone(), l)

    raw_node.get_raft().become_candidate()
    raw_node.get_raft().become_leader()

    rd = raw_node.ready()
    s.wl().append(rd.entries())
    raw_node.advance(rd.make_ref())

    data = [1] * 1000
    for _ in range(0, 10):
        raw_node.propose([], data)

    rd = raw_node.ready()
    entries = rd.entries()
    assert len(entries) == 10
    s.wl().append(entries)
    msgs = rd.messages()
    # election, and the first proposal (only one proposal gets sent
    # First append has two entries: the empty entry to confirm the
    # because we're in probe state).
    assert len(msgs) == 1
    assert msgs[0].get_msg_type() == MessageType.MsgAppend
    assert len(msgs[0].get_entries()) == 2
    raw_node.advance_append(rd.make_ref())

    s.wl().trigger_log_unavailable(True)

    # Become replicate state
    append_response = new_message(2, 1, MessageType.MsgAppendResponse, 0)
    append_response.set_term(2)
    append_response.set_index(2)
    raw_node.step(append_response)

    raw_node.get_raft().become_follower(raw_node.get_raft().get_term() + 1, 2)
    raw_node.get_raft().become_candidate()
    raw_node.get_raft().become_leader()

    # Entries are not sent due to the leadership or the term is changed.
    s.wl().trigger_log_unavailable(False)
    context = s.wl().take_get_entries_context()
    raw_node.on_entries_fetched(context.make_ref())
    rd = raw_node.ready()
    # no-op entry
    assert len(rd.entries()) == 1
    assert not rd.messages()
    raw_node.advance_append(rd.make_ref())


def test_raw_node_with_async_apply():
    l = default_logger()
    s = new_storage()
    s.wl().apply_snapshot(new_snapshot(1, 1, [1]))

    raw_node = new_raw_node(1, [1], 10, 1, s.clone(), l)
    raw_node.campaign()
    rd = raw_node.ready()

    # Single node should become leader.
    if ss := rd.ss():
        assert ss.get_leader_id() == raw_node.get_raft().get_leader_id()
    else:
        assert False

    s.wl().append(rd.entries())
    raw_node.advance(rd.make_ref())

    last_index = raw_node.get_raft().get_raft_log().last_index()

    data = b"hello world!"

    for _ in range(1, 10):
        cnt = random.randint(0, MAX_UINT64) % 10 + 1

        for _ in range(0, cnt):
            raw_node.propose([], data)

        rd = raw_node.ready()
        entries = list(map(lambda entry: entry.clone(), rd.entries()))
        assert entries[0].get_index() == last_index + 1
        assert entries[-1].get_index() == last_index + cnt
        must_cmp_ready(rd.make_ref(), None, None, entries, [], None, True, True, True)

        s.wl().append(entries)

        light_rd = raw_node.advance_append(rd.make_ref())
        assert entries == light_rd.committed_entries()
        assert light_rd.commit_index() == last_index + cnt

        # No matter how applied index changes, the index of next committed
        # entries should be the same.
        raw_node.advance_apply_to(last_index + 1)
        assert not raw_node.has_ready()

        last_index += cnt


# Test if the ready process is expected when a follower receives a snapshot
# and some committed entries after its snapshot.
def test_raw_node_entries_after_snapshot():
    l = default_logger()
    s = new_storage()
    s.wl().apply_snapshot(new_snapshot(1, 1, [1, 2]))

    raw_node = new_raw_node(1, [1, 2], 10, 1, s.clone(), l)

    entries = []
    for i in range(2, 20):
        entries.append(new_entry(2, i, "hello"))

    append_msg = new_message_with_entries(2, 1, MessageType.MsgAppend, entries)
    append_msg.set_term(2)
    append_msg.set_index(1)
    append_msg.set_log_term(1)
    append_msg.set_commit(5)
    raw_node.step(append_msg)

    rd = raw_node.ready()
    ss = soft_state(2, StateRole.Follower)

    must_cmp_ready(
        rd.make_ref(),
        ss.make_ref(),
        hard_state(2, 5, 0),
        entries,
        [],
        None,
        True,
        False,
        True,
    )

    s.wl().set_hardstate(rd.hs().clone())
    s.wl().append(rd.entries())
    light_rd = raw_node.advance(rd.make_ref())
    assert not light_rd.commit_index()
    assert light_rd.committed_entries() == entries[:4]
    assert not light_rd.messages()

    snapshot = new_snapshot(10, 3, [1, 2])
    snapshot_msg = new_message(2, 1, MessageType.MsgSnapshot, 0)
    snapshot_msg.set_term(3)
    snapshot_msg.set_snapshot(snapshot.clone())
    raw_node.step(snapshot_msg)

    entries = []
    for i in range(11, 14):
        entries.append(new_entry(3, i, "hello"))

    append_msg = new_message_with_entries(2, 1, MessageType.MsgAppend, entries)
    append_msg.set_term(3)
    append_msg.set_index(10)
    append_msg.set_log_term(3)
    append_msg.set_commit(12)
    raw_node.step(append_msg)

    rd = raw_node.ready()
    # If there is a snapshot, the committed entries should be empty.
    must_cmp_ready(
        rd.make_ref(),
        None,
        hard_state(3, 12, 0),
        entries,
        [],
        snapshot,
        True,
        False,
        True,
    )
    # Should have a MsgAppendResponse
    assert rd.persisted_messages()[0].get_msg_type() == MessageType.MsgAppendResponse

    s.wl().set_hardstate(rd.hs().clone())
    s.wl().apply_snapshot(rd.snapshot().clone())
    s.wl().append(rd.entries())

    light_rd = raw_node.advance(rd.make_ref())
    assert not light_rd.commit_index()
    assert light_rd.committed_entries() == entries[:2]
    assert not light_rd.messages()


# Test if the given committed entries are persisted when some persisted
# entries are overwritten by a new leader.
def test_raw_node_overwrite_entries():
    l = default_logger()
    s = new_storage()
    s.wl().apply_snapshot(new_snapshot(1, 1, [1, 2, 3]))

    raw_node = new_raw_node(1, [1, 2, 3], 10, 1, s.clone(), l)

    entries = [
        new_entry(2, 2, "hello"),
        new_entry(2, 3, "hello"),
        new_entry(2, 4, "hello"),
    ]
    append_msg = new_message_with_entries(2, 1, MessageType.MsgAppend, entries)
    append_msg.set_term(2)
    append_msg.set_index(1)
    append_msg.set_log_term(1)
    append_msg.set_commit(1)
    raw_node.step(append_msg)

    rd = raw_node.ready()
    ss = soft_state(2, StateRole.Follower)
    must_cmp_ready(
        rd.make_ref(),
        ss.make_ref(),
        hard_state(2, 1, 0),
        entries,
        [],
        None,
        True,
        False,
        True,
    )
    # Should have a MsgAppendResponse
    assert rd.persisted_messages()[0].get_msg_type() == MessageType.MsgAppendResponse
    s.wl().set_hardstate(rd.hs().clone())
    s.wl().append(rd.entries())

    light_rd = raw_node.advance(rd.make_ref())
    assert not light_rd.commit_index()
    assert not light_rd.committed_entries()
    assert not light_rd.messages()

    entries_2 = [
        new_entry(3, 4, "hello"),
        new_entry(3, 5, "hello"),
        new_entry(3, 6, "hello"),
    ]
    append_msg = new_message_with_entries(3, 1, MessageType.MsgAppend, entries_2)
    append_msg.set_term(3)
    append_msg.set_index(3)
    append_msg.set_log_term(2)
    append_msg.set_commit(5)
    raw_node.step(append_msg)

    rd = raw_node.ready()
    ss = soft_state(3, StateRole.Follower)
    must_cmp_ready(
        rd.make_ref(),
        ss.make_ref(),
        hard_state(3, 5, 0),
        entries_2,
        entries[:2],
        None,
        True,
        False,
        True,
    )
    # Should have a MsgAppendResponse
    assert rd.persisted_messages()[0].get_msg_type() == MessageType.MsgAppendResponse
    s.wl().set_hardstate(rd.hs().clone())
    s.wl().append(rd.entries())

    light_rd = raw_node.advance(rd.make_ref())
    assert not light_rd.commit_index()
    assert light_rd.committed_entries() == entries_2[:2]
    assert not light_rd.messages()


# Test if async ready process is expected when a leader receives
# the append response and persist its entries.
def test_async_ready_leader():
    l = default_logger()
    s = new_storage()
    s.wl().apply_snapshot(new_snapshot(1, 1, [1, 2, 3]))

    raw_node = new_raw_node(1, [1, 2, 3], 10, 1, s.clone(), l)
    raw_node.get_raft().become_candidate()
    raw_node.get_raft().become_leader()
    rd = raw_node.ready()

    if ss := rd.ss():
        assert ss.get_leader_id() == raw_node.get_raft().get_leader_id()
    else:
        assert False

    s.wl().append(rd.entries())
    raw_node.advance(rd.make_ref())

    assert raw_node.get_raft().get_term() == 2
    first_index = raw_node.get_raft().get_raft_log().last_index()

    data = b"hello world!"

    # Set node 2 progress to replicate
    raw_node.get_raft().prs().get(2).set_matched(1)
    raw_node.get_raft().prs().get(2).become_replicate()

    for i in range(0, 10):
        for _ in range(0, 10):
            raw_node.propose([], data)

        rd = raw_node.ready()
        assert rd.number() == i + 2
        entries = list(map(lambda entry: entry.clone(), rd.entries()))
        assert entries[0].get_index() == first_index + i * 10 + 1
        assert entries[-1].get_index() == first_index + i * 10 + 10

        # Leader‘s msg can be sent immediately.
        must_cmp_ready(rd.make_ref(), None, None, entries, [], None, False, True, True)
        for msg in rd.take_messages():
            assert msg.get_msg_type() == MessageType.MsgAppend

        s.wl().append(entries)
        raw_node.advance_append_async(rd.make_ref())

    # Unpersisted Ready number in range [2, 11]
    raw_node.on_persist_ready(4)
    # No new committed entries due to two nodes in this cluster
    assert not raw_node.has_ready()

    # The index of uncommitted entries in range [first_index, first_index + 100]
    append_response = new_message(2, 1, MessageType.MsgAppendResponse, 0)
    append_response.set_term(2)
    append_response.set_index(first_index + 100)

    raw_node.step(append_response)

    # Forward commit index due to append response
    rd = raw_node.ready()
    assert rd.hs() == hard_state(2, first_index + 30, 1)
    assert rd.committed_entries()[0].get_index() == first_index
    assert rd.committed_entries()[-1].get_index() == first_index + 30
    assert rd.messages()

    s.wl().set_hardstate(rd.hs().clone())
    raw_node.advance_append_async(rd.make_ref())

    # Forward commit index due to persist ready
    raw_node.on_persist_ready(8)
    rd = raw_node.ready()
    assert rd.hs() == hard_state(2, first_index + 70, 1)
    assert rd.committed_entries()[0].get_index() == first_index + 31
    assert rd.committed_entries()[-1].get_index() == first_index + 70
    assert rd.messages()
    assert not rd.persisted_messages()
    s.wl().set_hardstate(rd.hs().clone())

    # Forward commit index due to persist last ready
    light_rd = raw_node.advance_append(rd.make_ref())
    assert light_rd.commit_index() == first_index + 100
    assert light_rd.committed_entries()[0].get_index() == first_index + 71
    assert light_rd.committed_entries()[-1].get_index() == first_index + 100
    assert light_rd.messages()

    # Test when 2 followers response the append entries msg and leader has
    # not persisted them yet.
    first_index += 100
    for _ in range(0, 10):
        raw_node.propose([], data)

    rd = raw_node.ready()
    assert rd.number() == 14
    entries = list(map(lambda entry: entry.clone(), rd.entries()))
    assert entries[0].get_index(), first_index + 1
    assert entries[-1].get_index(), first_index + 10
    # Leader‘s msg can be sent immediately.
    must_cmp_ready(rd.make_ref(), None, None, entries, [], None, False, True, True)
    for msg in rd.take_messages():
        assert msg.get_msg_type() == MessageType.MsgAppend

    s.wl().append(entries)
    raw_node.advance_append_async(rd.make_ref())

    append_response = new_message(2, 1, MessageType.MsgAppendResponse, 0)
    append_response.set_term(2)
    append_response.set_index(first_index + 9)

    raw_node.step(append_response)

    append_response = new_message(3, 1, MessageType.MsgAppendResponse, 0)
    append_response.set_term(2)
    append_response.set_index(first_index + 10)

    raw_node.step(append_response)

    rd = raw_node.ready()
    # It should has some append msgs and its commit index should be first_index + 9.
    must_cmp_ready(
        rd.make_ref(),
        None,
        hard_state(2, first_index + 9, 1),
        [],
        [],
        None,
        False,
        True,
        False,
    )
    for msg in rd.take_messages():
        assert msg.get_msg_type() == MessageType.MsgAppend
        assert msg.get_commit() == first_index + 9

    # Forward commit index due to peer 1's append response and persisted entries
    light_rd = raw_node.advance_append(rd.make_ref())
    assert light_rd.commit_index() == first_index + 10
    assert light_rd.committed_entries()[0].get_index() == first_index + 1
    assert light_rd.committed_entries()[-1].get_index() == first_index + 10
    assert light_rd.messages()


# Test if async ready process is expected when a follower receives
# some append msg and snapshot.
def test_async_ready_follower():
    l = default_logger()
    s = new_storage()
    s.wl().apply_snapshot(new_snapshot(1, 1, [1, 2]))

    raw_node = new_raw_node(1, [1, 2], 10, 1, s.clone(), l)
    first_index = 1
    rd_number = 0
    for cnt in range(0, 3):
        for i in range(0, 10):
            entries = [
                new_entry(2, first_index + i * 3 + 1, "hello"),
                new_entry(2, first_index + i * 3 + 2, "hello"),
                new_entry(2, first_index + i * 3 + 3, "hello"),
            ]
            append_msg = new_message_with_entries(2, 1, MessageType.MsgAppend, entries)
            append_msg.set_term(2)
            append_msg.set_index(first_index + i * 3)
            if cnt == 0 and i == 0:
                append_msg.set_log_term(1)
            else:
                append_msg.set_log_term(2)

            append_msg.set_commit(first_index + i * 3 + 3)
            raw_node.step(append_msg)

            rd = raw_node.ready()
            assert rd.number() == rd_number + i + 1
            assert rd.hs() == hard_state(2, first_index + i * 3 + 3, 0)
            assert rd.entries() == entries
            assert rd.committed_entries() == []
            assert not rd.messages()
            assert (
                rd.persisted_messages()[0].get_msg_type()
                == MessageType.MsgAppendResponse
            )

            s.wl().set_hardstate(rd.hs())
            s.wl().append(rd.entries())
            raw_node.advance_append_async(rd.make_ref())

        # Unpersisted Ready number in range [1, 10]
        raw_node.on_persist_ready(rd_number + 4)
        rd = raw_node.ready()
        assert not rd.hs()
        assert rd.committed_entries()[0].get_index() == first_index + 1
        assert rd.committed_entries()[-1].get_index() == first_index + 3 * 3 + 3
        assert not rd.messages()
        assert not rd.persisted_messages()

        light_rd = raw_node.advance_append(rd.make_ref())
        assert not light_rd.commit_index()
        assert light_rd.committed_entries()[0].get_index() == first_index + 3 * 3 + 4
        assert light_rd.committed_entries()[-1].get_index() == first_index + 10 * 3
        assert not light_rd.messages()

        first_index += 10 * 3
        rd_number += 11

    snapshot = new_snapshot(first_index + 5, 2, [1, 2])
    snapshot_msg = new_message(2, 1, MessageType.MsgSnapshot, 0)
    snapshot_msg.set_term(2)
    snapshot_msg.set_snapshot(snapshot.clone())
    raw_node.step(snapshot_msg)

    rd = raw_node.ready()
    assert rd.number() == rd_number + 1
    must_cmp_ready(
        rd.make_ref(),
        None,
        hard_state(2, first_index + 5, 0),
        [],
        [],
        snapshot.clone(),
        True,
        False,
        True,
    )

    s.wl().set_hardstate(rd.hs().clone())
    s.wl().apply_snapshot(snapshot)
    s.wl().append(rd.entries())
    raw_node.advance_append_async(rd.make_ref())

    entries = []
    for i in range(1, 10):
        entries.append(new_entry(2, first_index + 5 + i, "hello"))

    append_msg = new_message_with_entries(2, 1, MessageType.MsgAppend, entries)
    append_msg.set_term(2)
    append_msg.set_index(first_index + 5)
    append_msg.set_log_term(2)
    append_msg.set_commit(first_index + 5 + 3)
    raw_node.step(append_msg)

    rd = raw_node.ready()
    assert rd.number() == rd_number + 2
    must_cmp_ready(
        rd.make_ref(),
        None,
        hard_state(2, first_index + 5 + 3, 0),
        entries,
        [],
        None,
        True,
        False,
        True,
    )
    s.wl().set_hardstate(rd.hs().clone())
    s.wl().append(rd.entries())
    raw_node.advance_append_async(rd.make_ref())

    raw_node.on_persist_ready(rd_number + 1)
    assert raw_node.get_raft().get_raft_log().get_persisted() == first_index + 5
    raw_node.advance_apply_to(first_index + 5)

    raw_node.on_persist_ready(rd_number + 2)

    rd = raw_node.ready()
    must_cmp_ready(
        rd.make_ref(),
        None,
        None,
        [],
        entries[:3],
        None,
        True,
        True,
        False,
    )


# Test if a new leader immediately sends all messages recorded before without
# persisting.
def test_async_ready_become_leader():
    l = default_logger()
    s = new_storage()
    s.wl().apply_snapshot(new_snapshot(5, 5, [1, 2, 3]))

    raw_node = new_raw_node(1, [1, 2, 3], 10, 1, s.clone(), l)
    for _ in range(1, raw_node.get_raft().election_timeout() * 2):
        raw_node.get_raft().tick_election()

    rd = raw_node.ready()
    assert rd.number() == 1

    ss = soft_state(0, StateRole.Candidate)
    must_cmp_ready(
        rd.make_ref(),
        ss.make_ref(),
        hard_state(6, 5, 1),
        [],
        [],
        None,
        True,
        False,
        True,
    )
    s.wl().set_hardstate(rd.hs().clone())

    for msg in rd.persisted_messages():
        assert msg.get_msg_type() == MessageType.MsgRequestVote

    raw_node.advance_append(rd.make_ref())

    # Peer 1 should reject to vote to peer 2
    vote_request_2 = new_message(2, 1, MessageType.MsgRequestVote, 0)
    vote_request_2.set_term(6)
    vote_request_2.set_log_term(4)
    vote_request_2.set_index(4)
    raw_node.step(vote_request_2)

    rd = raw_node.ready()
    assert rd.number() == 2
    must_cmp_ready(rd.make_ref(), None, None, [], [], None, True, False, False)
    assert (
        rd.persisted_messages()[0].get_msg_type() == MessageType.MsgRequestVoteResponse
    )
    raw_node.advance_append_async(rd.make_ref())

    # Peer 1 should reject to vote to peer 3
    vote_request_3 = new_message(3, 1, MessageType.MsgRequestVote, 0)
    vote_request_3.set_term(6)
    vote_request_3.set_log_term(4)
    vote_request_3.set_index(4)
    raw_node.step(vote_request_3)

    rd = raw_node.ready()
    assert rd.number() == 3
    must_cmp_ready(rd.make_ref(), None, None, [], [], None, True, False, False)
    assert (
        rd.persisted_messages()[0].get_msg_type() == MessageType.MsgRequestVoteResponse
    )

    raw_node.advance_append_async(rd.make_ref())

    # Peer 1 receives the vote from peer 2
    vote_response_2 = new_message(2, 1, MessageType.MsgRequestVoteResponse, 0)
    vote_response_2.set_term(6)
    vote_response_2.set_reject(False)
    raw_node.step(vote_response_2)

    rd = raw_node.ready()
    assert rd.number() == 4
    assert len(rd.entries()) == 1

    ss = soft_state(1, StateRole.Leader)
    must_cmp_ready(
        rd.make_ref(),
        ss.make_ref(),
        None,
        rd.entries(),
        [],
        None,
        False,
        True,
        True,
    )
    assert len(rd.messages()) == 2
    for msg in rd.take_messages():
        assert msg.get_msg_type() == MessageType.MsgAppend

    s.wl().append(rd.entries())
    s.wl().append(rd.entries())

    light_rd = raw_node.advance_append(rd.make_ref())
    assert not light_rd.commit_index()
    assert not light_rd.committed_entries()
    assert not light_rd.messages()


def test_async_ready_multiple_snapshot():
    l = default_logger()
    s = new_storage()
    s.wl().apply_snapshot(new_snapshot(1, 1, [1, 2]))

    raw_node = new_raw_node(1, [1, 2], 10, 1, s.clone(), l)

    snapshot = new_snapshot(10, 2, [1, 2])
    snapshot_msg = new_message(2, 1, MessageType.MsgSnapshot, 0)
    snapshot_msg.set_term(2)
    snapshot_msg.set_snapshot(snapshot.clone())
    raw_node.step(snapshot_msg)

    entries = []
    for i in range(11, 14):
        entries.append(new_entry(2, i, "hello"))

    append_msg = new_message_with_entries(2, 1, MessageType.MsgAppend, entries)
    append_msg.set_term(2)
    append_msg.set_index(10)
    append_msg.set_log_term(2)
    append_msg.set_commit(12)
    raw_node.step(append_msg)

    rd = raw_node.ready()
    assert rd.number() == 1
    # If there is a snapshot, the committed entries should be empty.
    ss = soft_state(2, StateRole.Follower)
    must_cmp_ready(
        rd.make_ref(),
        ss.make_ref(),
        hard_state(2, 12, 0),
        entries,
        [],
        snapshot,
        True,
        False,
        True,
    )
    s.wl().set_hardstate(rd.hs().clone())
    s.wl().apply_snapshot(rd.snapshot().clone())
    s.wl().append(rd.entries())

    raw_node.advance_append_async(rd.make_ref())

    snapshot = new_snapshot(20, 1, [1, 2])
    snapshot_msg = new_message(2, 1, MessageType.MsgSnapshot, 0)
    snapshot_msg.set_term(2)
    snapshot_msg.set_snapshot(snapshot.clone())
    raw_node.step(snapshot_msg)

    raw_node.on_persist_ready(1)

    assert raw_node.get_raft().get_raft_log().get_persisted() == 13

    raw_node.advance_apply_to(10)

    rd = raw_node.ready()
    assert rd.number() == 2
    must_cmp_ready(
        rd.make_ref(),
        None,
        hard_state(2, 20, 0),
        [],
        [],
        snapshot,
        True,
        False,
        True,
    )
    s.wl().set_hardstate(rd.hs().clone())
    s.wl().apply_snapshot(rd.snapshot().clone())

    light_rd = raw_node.advance_append(rd.make_ref())
    assert not light_rd.commit_index()
    assert not light_rd.committed_entries()
    assert not light_rd.messages()

    raw_node.advance_apply_to(20)


def test_committed_entries_pagination():
    l = default_logger()
    s = new_storage()
    raw_node = new_raw_node(1, [1, 2, 3], 10, 1, s, l)

    entries = []
    for i in range(2, 10):
        entries.append(new_entry(1, i, None))

    msg = new_message_with_entries(3, 1, MessageType.MsgAppend, entries)
    msg.set_term(1)
    msg.set_index(1)
    msg.set_log_term(1)
    msg.set_commit(9)
    raw_node.step(msg)

    # Test unpersisted entries won't be fetched.
    # NOTE: maybe it's better to allow fetching unpersisted committed entries.
    rd = raw_node.ready()
    assert not rd.committed_entries()
    assert raw_node.has_ready()

    # Persist entries.
    assert rd.entries()
    raw_node.store().wl().append(rd.entries())

    # Advance the ready, and we can get committed_entries as expected.
    # Test using 0 as `committed_entries_max_size` works as expected.
    raw_node.get_raft().set_max_committed_size_per_ready(0)
    rd = raw_node.advance(rd.make_ref())
    # `MemStorage::entries` uses `util::limit_size` to limit size of committed entries.
    # So there will be at least one entry.
    assert len(rd.committed_entries()) == 1

    # Fetch a `Ready` again without size limit for committed entries.
    assert raw_node.has_ready()
    raw_node.get_raft().set_max_committed_size_per_ready(MAX_UINT64)
    rd = raw_node.ready()
    assert len(rd.committed_entries()) == 7

    # No more `Ready`s.
    assert not raw_node.has_ready()


# Test with `commit_since_index`, committed entries can be fetched correctly after restart.
#
# Case steps:
# - Node learns that index 10 is committed
# - `next_entries` returns entries [2..11) in committed_entries (but index 10 already
#   exceeds maxBytes), which isn't noticed internally by Raft
# - Commit index gets bumped to 10
# - The node persists the `HardState`, but crashes before applying the entries
# - Upon restart, the storage returns the same entries, but `slice` takes a
#   different code path and removes the last entry.
# - Raft does not emit a HardState, but when the app calls advance(), it bumps
#   its internal applied index cursor to 10 (when it should be 9)
# - The next `Ready` asks the app to apply index 11 (omitting index 10), losing a
#   write.
def test_committed_entries_pagination_after_restart():
    class IgnoreSizeHintMemStorage:
        def __init__(self, store: MemStorage) -> None:
            self.inner = store

        def entries(self, low: int, high: int, _max_size: int) -> List[Entry]:
            return self.inner.entries(low, high, MAX_UINT64)

        def term(self, idx: int) -> int:
            return self.inner.term(idx)

        def first_index(self) -> int:
            return self.inner.first_index()

        def last_index(self) -> int:
            return self.inner.last_index()

        def snapshot(self, request_index: int) -> Snapshot:
            return self.inner.snapshot(request_index)

    l = default_logger()
    s = IgnoreSizeHintMemStorage(MemStorage.default())
    s.inner.wl().apply_snapshot(new_snapshot(1, 1, [1, 2, 3]))

    entries, size = ([], 0)
    for i in range(2, 10 + 1):
        e = new_entry(1, i, "test data")
        size += e.compute_size()
        entries.append(e)

    s.inner.wl().append(entries)
    s.inner.wl().hard_state().set_commit(10)

    s.inner.wl().append([new_entry(1, 11, "boom")])

    config = new_test_config(1, 10, 1)
    raw_node = InMemoryRawNode(config, s.inner, l)

    # `IgnoreSizeHintMemStorage` will ignore `max_committed_size_per_ready` but
    # `RaftLog::slice won't.`
    raw_node.get_raft().set_max_committed_size_per_ready(size - 1)

    highest_applied = 1
    while highest_applied != 11:
        rd = raw_node.ready()
        committed_entries = rd.take_committed_entries()
        next = committed_entries[0].get_index()
        assert highest_applied + 1 == next

        highest_applied = committed_entries[-1].get_index()
        raw_node.get_raft().get_raft_log().commit_to(11)



import pytest
from rraft import (
    ConfState,
    Logger_Ref,
    MemStorage,
    Config,
    InMemoryRaft,
    default_logger,
)


def new_storage(voters: int, learners: int) -> MemStorage:
    cc = ConfState.default()
    for i in range(1, voters + 1):
        cc.set_voters([*cc.get_voters(), i])
    for i in range(1, learners + 1):
        cc.set_learners([*cc.get_learners(), voters + i])

    return MemStorage.new_with_conf_state(cc)


def quick_raft(storage: MemStorage, logger: Logger_Ref):
    id = 1
    config = Config(id)
    return InMemoryRaft(config, storage, logger)


@pytest.mark.benchmark(group="raft-creation", warmup=True)
def test_raft_creation_0_0(benchmark):
    logger = default_logger()
    storage = new_storage(0, 0)
    benchmark(quick_raft, storage, logger)


@pytest.mark.benchmark(group="raft-creation", warmup=True)
def test_raft_creation_3_1(benchmark):
    logger = default_logger()
    storage = new_storage(3, 1)
    benchmark(quick_raft, storage, logger)


@pytest.mark.benchmark(group="raft-creation", warmup=True)
def test_raft_creation_5_2(benchmark):
    logger = default_logger()
    storage = new_storage(5, 2)
    benchmark(quick_raft, storage, logger)


@pytest.mark.benchmark(group="raft-creation", warmup=True)
def test_raft_creation_7_3(benchmark):
    logger = default_logger()
    storage = new_storage(7, 3)
    benchmark(quick_raft, storage, logger)


# ---


@pytest.mark.benchmark(group="raft-campaign", warmup=True)
def test_raft_campaign__3_1_CampaignPreElection(benchmark):
    logger = default_logger()
    storage = new_storage(3, 1)

    def bench(storage, logger):
        raft = quick_raft(storage, logger)
        raft.campaign("CampaignPreElection")

    benchmark(bench, storage, logger)


@pytest.mark.benchmark(group="raft-campaign", warmup=True)
def test_raft_campaign__3_1_CampaignElection(benchmark):
    logger = default_logger()
    storage = new_storage(3, 1)

    def bench(storage, logger):
        raft = quick_raft(storage, logger)
        raft.campaign("CampaignElection")

    benchmark(bench, storage, logger)


@pytest.mark.benchmark(group="raft-campaign", warmup=True)
def test_raft_campaign__3_1_CampaignTransfer(benchmark):
    logger = default_logger()
    storage = new_storage(3, 1)

    def bench(storage, logger):
        raft = quick_raft(storage, logger)
        raft.campaign("CampaignTransfer")

    benchmark(bench, storage, logger)


@pytest.mark.benchmark(group="raft-campaign", warmup=True)
def test_raft_campaign__5_2_CampaignPreElection(benchmark):
    logger = default_logger()
    storage = new_storage(5, 2)

    def bench(storage, logger):
        raft = quick_raft(storage, logger)
        raft.campaign("CampaignPreElection")

    benchmark(bench, storage, logger)


@pytest.mark.benchmark(group="raft-campaign", warmup=True)
def test_raft_campaign__5_2_CampaignElection(benchmark):
    logger = default_logger()
    storage = new_storage(5, 2)

    def bench(storage, logger):
        raft = quick_raft(storage, logger)
        raft.campaign("CampaignElection")

    benchmark(bench, storage, logger)


@pytest.mark.benchmark(group="raft-campaign", warmup=True)
def test_raft_campaign__5_2_CampaignTransfer(benchmark):
    logger = default_logger()
    storage = new_storage(5, 2)

    def bench(storage, logger):
        raft = quick_raft(storage, logger)
        raft.campaign("CampaignTransfer")

    benchmark(bench, storage, logger)


@pytest.mark.benchmark(group="raft-campaign", warmup=True)
def test_raft_campaign__7_3_CampaignPreElection(benchmark):
    logger = default_logger()
    storage = new_storage(7, 3)

    def bench(storage, logger):
        raft = quick_raft(storage, logger)
        raft.campaign("CampaignPreElection")

    benchmark(bench, storage, logger)


@pytest.mark.benchmark(group="raft-campaign", warmup=True)
def test_raft_campaign__7_3_CampaignElection(benchmark):
    logger = default_logger()
    storage = new_storage(7, 3)

    def bench(storage, logger):
        raft = quick_raft(storage, logger)
        raft.campaign("CampaignElection")

    benchmark(bench, storage, logger)


@pytest.mark.benchmark(group="raft-campaign", warmup=True)
def test_raft_campaign__7_3_CampaignTransfer(benchmark):
    logger = default_logger()
    storage = new_storage(7, 3)

    def bench(storage, logger):
        raft = quick_raft(storage, logger)
        raft.campaign("CampaignTransfer")

    benchmark(bench, storage, logger)



import pytest
from typing import Dict, List
from harness.src.interface import Interface
from harness.src.network import Network
from rraft import (
    ConfState,
    Entry,
    MemStorage,
    MemStorageRef,
    Message,
    MessageRef,
    MessageType,
    StateRole,
    default_logger,
    INVALID_ID,
)

from harness.utils import (
    ltoa,
    new_message,
    new_storage,
    new_test_raft,
    empty_entry,
    hard_state,
    new_entry,
    SOME_DATA,
    new_test_config,
    new_test_raft_with_config,
)


def commit_noop_entry(r: Interface, s: MemStorageRef):
    assert r.raft.get_state() == StateRole.Leader
    r.raft.bcast_append()

    # simulate the response of MsgAppend
    msgs = r.read_messages()
    for m in msgs:
        assert m.get_msg_type() == MessageType.MsgAppend
        assert len(m.get_entries()) == 1
        assert len(m.get_entries()[0].get_data()) == 0
        reply = accept_and_reply(m)
        r.step(reply)

    # ignore further messages to refresh followers' commit index
    r.read_messages()
    unstable = r.raft_log.unstable_entries()

    if unstable:
        e = unstable[-1]
        last_idx, last_term = e.get_index(), e.get_term()
        r.raft_log.stable_entries(last_idx, last_term)
        s.wl().append(unstable)
        r.raft.on_persist_entries(last_idx, last_term)
        committed = r.raft_log.get_committed()
        r.raft.commit_apply(committed)


def accept_and_reply(m: MessageRef) -> Message:
    assert m.get_msg_type() == MessageType.MsgAppend
    reply = new_message(m.get_to(), m.get_from(), MessageType.MsgAppendResponse, 0)
    reply.set_term(m.get_term())
    reply.set_index(m.get_index() + len(m.get_entries()))
    return reply


# test_update_term_from_message tests that if one server’s current term is
# smaller than the other’s, then it updates its current term to the larger
# value. If a candidate or leader discovers that its term is out of date,
# it immediately reverts to follower state.
# Reference: section 5.1
#
# test_follower_update_term_from_message
# test_candidate_update_term_from_message
# test_leader_update_term_from_message
@pytest.mark.parametrize(
    "state", [StateRole.Follower, StateRole.Candidate, StateRole.Leader]
)
def test_update_term_from_message(state: StateRole):
    l = default_logger()
    storage = new_storage()
    r = new_test_raft(1, [1, 2, 3], 10, 1, storage, l)

    if state == StateRole.Follower:
        r.raft.become_follower(1, 2)
    elif state == StateRole.PreCandidate:
        r.raft.become_pre_candidate()
    elif state == StateRole.Candidate:
        r.raft.become_candidate()
    elif state == StateRole.Leader:
        r.raft.become_candidate()
        r.raft.become_leader()
    else:
        assert False, "Invalid state"

    m = new_message(0, 0, MessageType.MsgAppend, 0)
    m.set_term(2)
    r.step(m)
    assert r.raft.get_term() == 2
    assert r.raft.get_state() == StateRole.Follower


# test_start_as_follower tests that when servers start up, they begin as followers.
# Reference: section 5.2
def test_start_as_follower():
    l = default_logger()
    storage = new_storage()
    r = new_test_raft(1, [1, 2, 3], 10, 1, storage, l)
    assert r.raft.get_state() == StateRole.Follower


# test_leader_bcast_beat tests that if the leader receives a heartbeat tick,
# it will send a msgApp with m.Index = 0, m.LogTerm=0 and empty entries as
# heartbeat to all followers.
# Reference: section 5.2
def test_leader_bcast_beat():
    l = default_logger()
    storage = new_storage()
    hi = 1
    r = new_test_raft(1, [1, 2, 3], 10, 1, storage, l)
    r.raft.become_candidate()
    r.raft.become_leader()
    for i in range(0, 10):
        ent = empty_entry(0, i + 1)
        r.raft.append_entry([ent])

    for _ in range(0, hi):
        r.raft.tick()

    msgs = r.read_messages()
    msgs.sort(key=lambda m: str(m))

    def new_message_ext(f: int, to: int) -> Message:
        m = new_message(f, to, MessageType.MsgHeartbeat, 0)
        m.set_term(1)
        m.set_commit(0)
        return m

    expect_msgs = [new_message_ext(1, 2), new_message_ext(1, 3)]
    assert msgs == expect_msgs


# test_nonleader_start_election tests that if a follower receives no communication
# over election timeout, it begins an election to choose a new leader. It
# increments its current term and transitions to candidate state. It then
# votes for itself and issues RequestVote RPCs in parallel to each of the
# other servers in the cluster.
# Reference: section 5.2
# Also if a candidate fails to obtain a majority, it will time out and
# start a new election by incrementing its term and initiating another
# round of RequestVote RPCs.
# Reference: section 5.2
#
# test_follower_start_election
# test_candidate_start_new_election
@pytest.mark.parametrize("state", [StateRole.Follower, StateRole.Candidate])
def test_nonleader_start_election(state: StateRole):
    l = default_logger()
    storage = new_storage()
    # election timeout
    et = 10

    r = new_test_raft(1, [1, 2, 3], et, 1, storage, l)

    if state == StateRole.Follower:
        r.raft.become_follower(1, 2)
    elif state == StateRole.Candidate:
        r.raft.become_candidate()
    else:
        assert False, "Only non-leader role is accepted."

    for _ in range(1, 2 * et):
        r.raft.tick()

    assert r.raft.get_term() == 2
    assert r.raft.get_state() == StateRole.Candidate
    assert r.raft.prs().votes()[r.raft.get_id()]

    msgs = r.read_messages()
    msgs.sort(key=lambda m: str(m))

    def new_message_ext(f: int, to: int) -> Message:
        m = new_message(f, to, MessageType.MsgRequestVote, 0)
        m.set_term(2)
        return m

    expect_msgs = [new_message_ext(1, 2), new_message_ext(1, 3)]
    assert msgs == expect_msgs


# test_leader_election_in_one_round_rpc tests all cases that may happen in
# leader election during one round of RequestVote RPC:
# a) it wins the election
# b) it loses the election
# c) it is unclear about the result
# Reference: section 5.2
def test_leader_election_in_one_round_rpc():
    l = default_logger()

    class Test:
        def __init__(self, size: int, votes: Dict[int, bool], state: StateRole):
            self.size = size
            self.votes = votes
            self.state = state

    tests = [
        # win the election when receiving votes from a majority of the servers
        Test(1, {}, StateRole.Leader),
        Test(3, {2: True, 3: True}, StateRole.Leader),
        Test(3, {2: True}, StateRole.Leader),
        Test(5, {2: True, 3: True, 4: True, 5: True}, StateRole.Leader),
        Test(5, {2: True, 3: True, 4: True}, StateRole.Leader),
        Test(5, {2: True, 3: True}, StateRole.Leader),
        # return to follower state if it receives vote denial from a majority
        Test(3, {2: False, 3: False}, StateRole.Follower),
        Test(5, {2: False, 3: False, 4: False, 5: False}, StateRole.Follower),
        Test(5, {2: True, 3: False, 4: False, 5: False}, StateRole.Follower),
        # stay in candidate if it does not obtain the majority
        Test(3, {}, StateRole.Candidate),
        Test(5, {2: True}, StateRole.Candidate),
        Test(5, {2: False, 2: False}, StateRole.Candidate),
        Test(5, {}, StateRole.Candidate),
    ]

    for i, v in enumerate(tests):
        size, votes, state = v.size, v.votes, v.state
        storage = new_storage()
        r = new_test_raft(1, list(range(1, size + 1)), 10, 1, storage, l)
        r.step(new_message(1, 1, MessageType.MsgHup, 0))

        for id, vote in votes.items():
            m = new_message(id, 1, MessageType.MsgRequestVoteResponse, 0)
            m.set_term(r.raft.get_term())
            m.set_reject(not vote)
            r.step(m)

        assert (
            r.raft.get_state() == state
        ), f"#{i}: state = {r.raft.get_state()}, want {state}"

        assert r.raft.get_term() == 1, f"#{i}: term = {r.raft.get_term()}, want {1}"


# test_follower_vote tests that each follower will vote for at most one
# candidate in a given term, on a first-come-first-served basis.
# Reference: section 5.2
def test_follower_vote():
    l = default_logger()

    class Test:
        def __init__(self, vote: int, nvote: int, wreject: bool):
            self.vote = vote
            self.nvote = nvote
            self.wreject = wreject

    tests = [
        Test(INVALID_ID, 1, False),
        Test(INVALID_ID, 2, False),
        Test(1, 1, False),
        Test(2, 2, False),
        Test(1, 2, True),
        Test(2, 1, True),
    ]

    for i, v in enumerate(tests):
        vote, nvote, wreject = v.vote, v.nvote, v.wreject
        storage = new_storage()
        r = new_test_raft(1, [1, 2, 3], 10, 1, storage, l)
        hs = hard_state(1, 0, vote)
        r.raft.load_state(hs)

        m = new_message(nvote, 1, MessageType.MsgRequestVote, 0)
        m.set_term(1)
        r.step(m)

        msgs = r.read_messages()
        m = new_message(1, nvote, MessageType.MsgRequestVoteResponse, 0)
        m.set_term(1)
        m.set_reject(wreject)
        expected_msgs = [m]
        assert msgs == expected_msgs, f"#{i}: msgs = {msgs}, want {expected_msgs}"


# test_candidate_fallback tests that while waiting for votes,
# if a candidate receives an AppendEntries RPC from another server claiming
# to be leader whose term is at least as large as the candidate's current term,
# it recognizes the leader as legitimate and returns to follower state.
# Reference: section 5.2
def test_candidate_fallback():
    l = default_logger()

    def new_message_ext(f: int, to: int, term: int) -> Message:
        m = new_message(f, to, MessageType.MsgAppend, 0)
        m.set_term(term)
        return m

    tests: List[Message] = [
        new_message_ext(2, 1, 2),
        new_message_ext(2, 1, 3),
    ]

    for i, m in enumerate(tests):
        storage = new_storage()
        r = new_test_raft(1, [1, 2, 3], 10, 1, storage, l)
        r.raft.step(new_message(1, 1, MessageType.MsgHup, 0))
        assert r.raft.get_state() == StateRole.Candidate

        term = m.get_term()
        r.raft.step(m)

        assert (
            r.raft.get_state() == StateRole.Follower
        ), f"#{i}: state = {r.raft.get_state()}, want {StateRole.Follower}"

        assert (
            r.raft.get_term() == term
        ), f"#{i}: term = {r.raft.get_term()}, want {term}"


# test_non_leader_election_timeout_randomized tests that election timeout for
# follower or candidate is randomized.
# Reference: section 5.2
#
# test_follower_election_timeout_randomized
# test_candidate_election_timeout_randomized
@pytest.mark.parametrize("state", [StateRole.Follower, StateRole.Candidate])
def test_non_leader_election_timeout_randomized(state: StateRole):
    l = default_logger()
    et = 10
    storage = new_storage()
    r = new_test_raft(1, [1, 2, 3], et, 1, storage, l)
    timeouts = {}
    for _ in range(1000 * et):
        term = r.raft.get_term()
        if state == StateRole.Follower:
            r.raft.become_follower(term + 1, 2)
        elif state == StateRole.Candidate:
            r.raft.become_candidate()
        else:
            assert False, "only non leader state is accepted!"

        time = 0
        while not r.read_messages():
            r.raft.tick()
            time += 1

        timeouts[time] = True


# test_nonleaders_election_timeout_nonconfict tests that in most cases only a
# single server(follower or candidate) will time out, which reduces the
# likelihood of split vote in the new election.
# Reference: section 5.2
#
# test_follower_election_timeout_nonconflict
# test_candidates_election_timeout_nonconf
@pytest.mark.parametrize("state", [StateRole.Follower, StateRole.Candidate])
def test_nonleaders_election_timeout_nonconfict(state: StateRole):
    l = default_logger()
    et = 10
    size = 5
    rs: List[Interface] = []
    ids = list(range(1, size + 1))
    for id in ids:
        storage = new_storage()
        rs.append(new_test_raft(id, ids, et, 1, storage, l))

    conflicts = 0

    for _ in range(0, 1000):
        for r in rs:
            term = r.raft.get_term()
            if state == StateRole.Follower:
                r.raft.become_follower(term + 1, INVALID_ID)
            elif state == StateRole.Candidate:
                r.raft.become_candidate()
            else:
                assert False, "non leader state is expect!"

        timeout_num = 0
        while timeout_num == 0:
            for r in rs:
                r.raft.tick()

                if not r.read_messages():
                    timeout_num += 1

        # several rafts time out at the same tick
        if timeout_num > 1:
            conflicts += 1


# test_leader_start_replication tests that when receiving client proposals,
# the leader appends the proposal to its log as a new entry, then issues
# AppendEntries RPCs in parallel to each of the other servers to replicate
# the entry. Also, when sending an AppendEntries RPC, the leader includes
# the index and term of the entry in its log that immediately precedes
# the new entries.
# Also, it writes the new entry into stable storage.
# Reference: section 5.3
def test_leader_start_replication():
    l = default_logger()
    s = new_storage()
    r = new_test_raft(1, [1, 2, 3], 10, 1, s, l)
    r.raft.become_candidate()
    r.raft.become_leader()
    commit_noop_entry(r, s)
    li = r.raft_log.last_index()

    r.step(new_message(1, 1, MessageType.MsgPropose, 1))

    assert r.raft_log.last_index() == li + 1
    assert r.raft_log.get_committed() == li

    msgs = r.read_messages()
    msgs.sort(key=lambda m: str(m))
    wents1, wents2 = [new_entry(1, li + 1, SOME_DATA)], [
        new_entry(1, li + 1, SOME_DATA)
    ]

    def new_message_ext(f: int, to: int, ents: List[Entry]) -> Message:
        m = new_message(f, to, MessageType.MsgAppend, 0)
        m.set_term(1)
        m.set_index(li)
        m.set_log_term(1)
        m.set_commit(li)
        m.set_entries(ents)
        return m

    excepted_msgs = [
        new_message_ext(1, 2, wents1),
        new_message_ext(1, 3, wents2),
    ]

    assert msgs == excepted_msgs
    assert r.raft_log.unstable_entries() == wents1


# test_leader_commit_entry tests that when the entry has been safely replicated,
# the leader gives out the applied entries, which can be applied to its state
# machine.
# Also, the leader keeps track of the highest index it knows to be committed,
# and it includes that index in future AppendEntries RPCs so that the other
# servers eventually find out.
# Reference: section 5.3
def test_leader_commit_entry():
    l = default_logger()
    s = new_storage()
    r = new_test_raft(1, [1, 2, 3], 10, 1, s, l)
    r.raft.become_candidate()
    r.raft.become_leader()
    commit_noop_entry(r, s)
    li = r.raft_log.last_index()

    propose = new_message(1, 1, MessageType.MsgPropose, 1)

    r.step(propose)
    r.persist()

    for m in r.read_messages():
        reply = accept_and_reply(m)
        r.step(reply)

    assert r.raft_log.get_committed() == li + 1

    wents = [new_entry(1, li + 1, SOME_DATA)]

    assert r.raft_log.next_entries(None) == wents

    msgs = r.read_messages()
    msgs.sort(key=lambda m: str(m))

    for i, m in enumerate(msgs):
        assert i + 2 == m.get_to()
        assert m.get_msg_type() == MessageType.MsgAppend
        assert m.get_commit() == li + 1


# test_leader_acknowledge_commit tests that a log entry is committed once the
# leader that created the entry has replicated it on a majority of the servers.
# Reference: section 5.3
def test_leader_acknowledge_commit():
    l = default_logger()

    class Test:
        def __init__(self, size: int, acceptors: Dict[int, bool], wack: bool):
            self.size = size
            self.acceptors = acceptors
            self.wack = wack

    tests = [
        Test(1, {}, True),
        Test(3, {}, False),
        Test(3, {2: True}, True),
        Test(3, {2: True, 3: True}, True),
        Test(5, {}, False),
        Test(5, {2: True}, False),
        Test(5, {2: True, 3: True}, True),
        Test(5, {2: True, 3: True, 4: True}, True),
        Test(5, {2: True, 3: True, 4: True, 5: True}, True),
    ]

    for i, v in enumerate(tests):
        size, acceptors, wack = v.size, v.acceptors, v.wack
        s = new_storage()
        r = new_test_raft(1, list(range(1, size + 1)), 10, 1, s, l)
        r.raft.become_candidate()
        r.raft.become_leader()
        commit_noop_entry(r, s)
        li = r.raft_log.last_index()
        r.step(new_message(1, 1, MessageType.MsgPropose, 1))
        r.persist()

        for m in r.read_messages():
            if m.get_to() in acceptors.keys():
                r.step(accept_and_reply(m))

        g = r.raft_log.get_committed() > li
        assert not (g ^ wack), f"#{i}: ack commit = {g}, want {wack}"


# test_leader_commit_preceding_entries tests that when leader commits a log entry,
# it also commits all preceding entries in the leader’s log, including
# entries created by previous leaders.
# Also, it applies the entry to its local state machine (in log order).
# Reference: section 5.3
def test_leader_commit_preceding_entries():
    l = default_logger()
    tests: List[List[Entry]] = [
        [],
        [empty_entry(2, 1)],
        [empty_entry(1, 1), empty_entry(2, 2)],
        [empty_entry(1, 1)],
    ]

    for i, tt in enumerate(tests):
        cs = ConfState([1, 2, 3], [])
        store = MemStorage.new_with_conf_state(cs)
        store.wl().append(tt)
        cfg = new_test_config(1, 10, 1)
        r = new_test_raft_with_config(cfg, store, l)

        hs = hard_state(2, 0, 0)
        r.raft.load_state(hs)
        r.raft.become_candidate()
        r.raft.become_leader()

        r.step(new_message(1, 1, MessageType.MsgPropose, 1))
        r.persist()

        for m in r.read_messages():
            r.step(accept_and_reply(m))

        li = len(tt)
        tt.extend([empty_entry(3, li + 1), new_entry(3, li + 2, SOME_DATA)])
        g = r.raft_log.next_entries(None)
        wg = tt

        assert g == wg, f"#{i}: ents = {g}, want {wg}"


# test_follower_commit_entry tests that once a follower learns that a log entry
# is committed, it applies the entry to its local state machine (in log order).
# Reference: section 5.3
def test_follower_commit_entry():
    l = default_logger()

    class Test:
        def __init__(self, ents: List[Entry], commit: int):
            self.ents = ents
            self.commit = commit

    tests = [
        Test([new_entry(1, 1, SOME_DATA)], 1),
        Test(
            [
                new_entry(1, 1, SOME_DATA),
                new_entry(1, 2, "somedata2"),
            ],
            2,
        ),
        Test(
            [
                new_entry(1, 1, "somedata2"),
                new_entry(1, 2, SOME_DATA),
            ],
            2,
        ),
        Test(
            [
                new_entry(1, 1, SOME_DATA),
                new_entry(1, 2, "somedata2"),
            ],
            1,
        ),
    ]

    for i, v in enumerate(tests):
        ents, commit = v.ents, v.commit
        storage = new_storage()
        r = new_test_raft(1, [1, 2, 3], 10, 1, storage, l)
        r.raft.become_follower(1, 2)

        m = new_message(2, 1, MessageType.MsgAppend, 0)
        m.set_term(1)
        m.set_commit(commit)
        m.set_entries(ents)
        r.step(m)
        r.persist()

        assert (
            r.raft_log.get_committed() == commit
        ), f"#{i}: committed = {r.raft_log.get_committed()}, want {commit}"

        wents = ents[:commit]
        g = r.raft_log.next_entries(None)
        assert g == wents, f"#{i}: next_ents = {g}, want {wents}"


# test_follower_check_msg_append tests that if the follower does not find an
# entry in its log with the same index and term as the one in AppendEntries RPC,
# then it refuses the new entries. Otherwise it replies that it accepts the
# append entries.
# Reference: section 5.3
def test_follower_check_msg_append():
    l = default_logger()
    ents = [empty_entry(1, 1), empty_entry(2, 2)]

    class Test:
        def __init__(
            self,
            term: int,
            index: int,
            windex: int,
            w_commit: int,
            wreject: bool,
            wreject_hint: int,
            w_log_term: int,
        ) -> None:
            self.term = term
            self.index = index
            self.windex = windex
            self.w_commit = w_commit
            self.wreject = wreject
            self.wreject_hint = wreject_hint
            self.w_log_term = w_log_term

    tests = [
        # match with committed entries
        Test(
            0,
            0,
            1,
            1,
            False,
            0,
            0,
        ),
        Test(
            ents[0].get_term(),
            ents[0].get_index(),
            1,
            1,
            False,
            0,
            0,
        ),
        # match with uncommitted entries
        Test(
            ents[1].get_term(),
            ents[1].get_index(),
            2,
            1,
            False,
            0,
            0,
        ),
        # unmatch with existing entry
        Test(
            ents[0].get_term(),
            ents[1].get_index(),
            ents[1].get_index(),
            1,
            True,
            1,
            1,
        ),
        # unexisting entry
        Test(
            ents[1].get_term() + 1,
            ents[1].get_index() + 1,
            ents[1].get_index() + 1,
            1,
            True,
            2,
            2,
        ),
    ]

    for i, v in enumerate(tests):
        term, index, windex, w_commit, wreject, wreject_hint, w_log_term = (
            v.term,
            v.index,
            v.windex,
            v.w_commit,
            v.wreject,
            v.wreject_hint,
            v.w_log_term,
        )

        cs = ConfState([1, 2, 3], [])
        store: MemStorage = MemStorage.new_with_conf_state(cs)
        store.wl().append(ents)
        cfg = new_test_config(1, 10, 1)
        r = new_test_raft_with_config(cfg, store, l)

        hs = hard_state(0, 1, 0)
        r.raft.load_state(hs)
        r.raft.become_follower(2, 2)

        m = new_message(2, 1, MessageType.MsgAppend, 0)
        m.set_term(2)
        m.set_log_term(term)
        m.set_index(index)
        r.step(m)

        msgs = r.read_messages()
        wm = new_message(1, 2, MessageType.MsgAppendResponse, 0)
        wm.set_term(2)
        wm.set_index(windex)
        wm.set_commit(w_commit)

        if wreject:
            wm.set_reject(wreject)
            wm.set_reject_hint(wreject_hint)
            wm.set_log_term(w_log_term)

        excepted_msgs = [wm]
        assert msgs == excepted_msgs, f"#{i}: msgs = {msgs}, want {excepted_msgs}"


# test_follower_append_entries tests that when AppendEntries RPC is valid,
# the follower will delete the existing conflict entry and all that follow it,
# and append any new entries not already in the log.
# Also, it writes the new entry into stable storage.
# Reference: section 5.3
def test_follower_append_entries():
    l = default_logger()

    class Test:
        def __init__(
            self,
            index: int,
            term: int,
            ents: List[Entry],
            wents: List[Entry],
            wunstable: List[Entry],
        ) -> None:
            self.index = index
            self.term = term
            self.ents = ents
            self.wents = wents
            self.wunstable = wunstable

    tests = [
        Test(
            2,
            2,
            [empty_entry(3, 3)],
            [empty_entry(1, 1), empty_entry(2, 2), empty_entry(3, 3)],
            [empty_entry(3, 3)],
        ),
        Test(
            1,
            1,
            [empty_entry(3, 2), empty_entry(4, 3)],
            [empty_entry(1, 1), empty_entry(3, 2), empty_entry(4, 3)],
            [empty_entry(3, 2), empty_entry(4, 3)],
        ),
        Test(0, 0, [empty_entry(1, 1)], [empty_entry(1, 1), empty_entry(2, 2)], []),
        Test(0, 0, [empty_entry(3, 1)], [empty_entry(3, 1)], [empty_entry(3, 1)]),
    ]

    for i, v in enumerate(tests):
        index, term, ents, wents, wunstable = (
            v.index,
            v.term,
            v.ents,
            v.wents,
            v.wunstable,
        )
        cs = ConfState([1, 2, 3], [])
        store = MemStorage.new_with_conf_state(cs)
        store.wl().append([empty_entry(1, 1), empty_entry(2, 2)])
        cfg = new_test_config(1, 10, 1)
        r = new_test_raft_with_config(cfg, store, l)
        r.raft.become_follower(2, 2)

        m = new_message(2, 1, MessageType.MsgAppend, 0)
        m.set_term(2)
        m.set_log_term(term)
        m.set_index(index)
        m.set_entries(ents)
        r.step(m)

        g = r.raft_log.all_entries()
        assert g == wents, f"#{i}: ents = {g}, want {wents}"

        g = r.raft_log.unstable_entries()
        assert g == wunstable, f"#{i}: unstable_entries = {g}, want {wunstable}"


# test_leader_sync_follower_log tests that the leader could bring a follower's log
# into consistency with its own.
# Reference: section 5.3, figure 7
def test_leader_sync_follower_log():
    l = default_logger()
    ents = [
        empty_entry(1, 1),
        empty_entry(1, 2),
        empty_entry(1, 3),
        empty_entry(4, 4),
        empty_entry(4, 5),
        empty_entry(5, 6),
        empty_entry(5, 7),
        empty_entry(6, 8),
        empty_entry(6, 9),
        empty_entry(6, 10),
    ]

    term = 8
    tests = [
        [
            empty_entry(1, 1),
            empty_entry(1, 2),
            empty_entry(1, 3),
            empty_entry(4, 4),
            empty_entry(4, 5),
            empty_entry(5, 6),
            empty_entry(5, 7),
            empty_entry(6, 8),
            empty_entry(6, 9),
        ],
        [
            empty_entry(1, 1),
            empty_entry(1, 2),
            empty_entry(1, 3),
            empty_entry(4, 4),
        ],
        [
            empty_entry(1, 1),
            empty_entry(1, 2),
            empty_entry(1, 3),
            empty_entry(4, 4),
            empty_entry(4, 5),
            empty_entry(5, 6),
            empty_entry(5, 7),
            empty_entry(6, 8),
            empty_entry(6, 9),
            empty_entry(6, 10),
            empty_entry(6, 11),
        ],
        [
            empty_entry(1, 1),
            empty_entry(1, 2),
            empty_entry(1, 3),
            empty_entry(4, 4),
            empty_entry(4, 5),
            empty_entry(5, 6),
            empty_entry(5, 7),
            empty_entry(6, 8),
            empty_entry(6, 9),
            empty_entry(6, 10),
            empty_entry(7, 11),
            empty_entry(7, 12),
        ],
        [
            empty_entry(1, 1),
            empty_entry(1, 2),
            empty_entry(1, 3),
            empty_entry(4, 4),
            empty_entry(4, 5),
            empty_entry(4, 6),
            empty_entry(4, 7),
        ],
        [
            empty_entry(1, 1),
            empty_entry(1, 2),
            empty_entry(1, 3),
            empty_entry(2, 4),
            empty_entry(2, 5),
            empty_entry(2, 6),
            empty_entry(3, 7),
            empty_entry(3, 8),
            empty_entry(3, 9),
            empty_entry(3, 10),
            empty_entry(3, 11),
        ],
    ]

    for i, tt in enumerate(tests):
        lead_cs = ConfState([1, 2, 3], [])
        lead_store = MemStorage.new_with_conf_state(lead_cs)
        lead_store.wl().append(ents)
        lead_cfg = new_test_config(1, 10, 1)
        lead = new_test_raft_with_config(lead_cfg, lead_store, l)
        last_index = lead.raft_log.last_index()
        lead_hs = hard_state(term, last_index, 0)
        lead.raft.load_state(lead_hs)

        follower_cs = ConfState([1, 2, 3], [])
        follower_store = MemStorage.new_with_conf_state(follower_cs)
        follower_store.wl().append(tt)

        follower_cfg = new_test_config(2, 10, 1)
        follower = new_test_raft_with_config(follower_cfg, follower_store, l)
        follower_hs = hard_state(term - 1, 0, 0)
        follower.raft.load_state(follower_hs)

        NOP_STEPPER = Interface(None)

        # It is necessary to have a three-node cluster.
        # The second may have more up-to-date log than the first one, so the
        # first node needs the vote from the third node to become the leader.
        n = Network.new([lead, follower, NOP_STEPPER], l)
        m = new_message(1, 1, MessageType.MsgHup, 0)
        n.send([m])

        # The election occurs in the term after the one we loaded with
        # lead.load_state above.
        m = new_message(3, 1, MessageType.MsgRequestVoteResponse, 0)
        m.set_term(term + 1)
        n.send([m])

        m = new_message(1, 1, MessageType.MsgPropose, 0)
        e = Entry.default()
        m.set_entries([e])
        n.send([m])

        lead_str = ltoa(n.peers.get(1).raft_log)
        follower_str = ltoa(n.peers.get(2).raft_log)

        assert (
            lead_str == follower_str
        ), f"#{i}: lead str: {lead_str}, follower_str: {follower_str}"


# test_vote_request tests that the vote request includes information about the candidate’s log
# and are sent to all of the other nodes.
# Reference: section 5.4.1
def test_vote_request():
    l = default_logger()

    class Test:
        def __init__(self, ents: List[Entry], wterm: int) -> None:
            self.ents = ents
            self.wterm = wterm

    tests = [
        Test([empty_entry(1, 1)], 2),
        Test([empty_entry(1, 1), empty_entry(2, 2)], 3),
    ]

    for j, v in enumerate(tests):
        ents, wterm = (v.ents, v.wterm)

        storage = new_storage()
        r = new_test_raft(1, [1, 2, 3], 10, 1, storage, l)
        m = new_message(2, 1, MessageType.MsgAppend, 0)
        m.set_term(wterm - 1)
        m.set_log_term(0)
        m.set_index(0)
        cloned_ents = list(map(lambda x: x.clone(), ents))
        m.set_entries(cloned_ents)
        r.step(m)
        r.read_messages()

        for _ in range(1, r.raft.election_timeout() * 2):
            r.raft.tick_election()

        msgs = r.read_messages()
        msgs.sort(key=lambda m: str(m))
        assert len(msgs) == 2, f"#{j}: msg count = {len(msgs)}, want 2"

        for i, m in enumerate(msgs):
            assert (
                m.get_msg_type() == MessageType.MsgRequestVote
            ), f"#{j}.{i}: msg_type = {m.get_msg_type()}, want {MessageType.MsgRequestVote}"

            assert m.get_to() == i + 2, f"#{j}.{i}: to = {m.get_to()}, want {i + 2}"

            assert (
                m.get_term() == wterm
            ), f"#{j}.{i}: term = {m.get_term()}, want {wterm}"

            windex = ents[-1].get_index()
            wlogterm = ents[-1].get_term()

            assert (
                m.get_index() == windex
            ), f"#{j}.{i}: index = {m.get_index()}, want {windex}"

            assert (
                m.get_log_term() == wlogterm
            ), f"#{j}.{i}: log_term = {m.get_log_term()}, want {wlogterm}"


# test_voter tests the voter denies its vote if its own log is more up-to-date
# than that of the candidate.
# Reference: section 5.4.1
def test_voter():
    l = default_logger()

    class Test:
        def __init__(
            self, ents: List[Entry], log_term: int, index: int, wreject: bool
        ) -> None:
            self.ents = ents
            self.log_term = log_term
            self.index = index
            self.wreject = wreject

    tests = [
        # same logterm
        Test([empty_entry(1, 1)], 1, 1, False),
        Test([empty_entry(1, 1)], 1, 2, False),
        Test([empty_entry(1, 1), empty_entry(1, 2)], 1, 1, True),
        # candidate higher logterm
        Test([empty_entry(1, 1)], 2, 1, False),
        Test([empty_entry(1, 1)], 1, 2, False),
        Test([empty_entry(1, 1), empty_entry(1, 2)], 2, 1, False),
        # voter higher logterm
        Test([empty_entry(2, 1)], 1, 1, True),
        Test([empty_entry(2, 1)], 1, 2, True),
        Test([empty_entry(2, 1), empty_entry(1, 2)], 1, 1, True),
    ]

    for i, v in enumerate(tests):
        ents, log_term, index, wreject = (v.ents, v.log_term, v.index, v.wreject)
        cs = ConfState([1, 2], [])
        s = MemStorage.new_with_conf_state(cs)
        s.wl().append(ents)
        cfg = new_test_config(1, 10, 1)
        r = new_test_raft_with_config(cfg, s, l)

        m = new_message(2, 1, MessageType.MsgRequestVote, 0)
        m.set_term(3)
        m.set_log_term(log_term)
        m.set_index(index)
        r.step(m)

        msgs = r.read_messages()
        assert len(msgs) == 1, f"#{i}: msg count = {len(msgs)}, want 1"

        assert (
            msgs[0].get_msg_type() == MessageType.MsgRequestVoteResponse
        ), f"#{i}: msg_type = {msgs[0].get_msg_type()}, want {MessageType.MsgRequestVoteResponse}"

        assert (
            msgs[0].get_reject() == wreject
        ), f"#{i}: reject = {msgs[0].get_reject()}, want {wreject}"


# TestLeaderOnlyCommitsLogFromCurrentTerm tests that only log entries from the leader’s
# current term are committed by counting replicas.
# Reference: section 5.4.2
def test_leader_only_commits_log_from_current_term():
    l = default_logger()
    ents = [empty_entry(1, 1), empty_entry(2, 2)]

    class Test:
        def __init__(self, index: int, wcommit: int) -> None:
            self.index = index
            self.wcommit = wcommit

    tests = [
        # do not commit log entries in previous terms
        Test(1, 0),
        Test(2, 0),
        # commit log in current term
        Test(3, 3),
    ]

    for i, v in enumerate(tests):
        index, wcommit = (v.index, v.wcommit)
        cs = ConfState([1, 2], [])
        store = MemStorage.new_with_conf_state(cs)
        store.wl().append(ents)
        cfg = new_test_config(1, 10, 1)
        r = new_test_raft_with_config(cfg, store, l)

        hs = hard_state(2, 0, 0)
        r.raft.load_state(hs)

        # become leader at term 3
        r.raft.become_candidate()
        r.raft.become_leader()
        r.read_messages()

        # propose a entry to current term
        r.step(new_message(1, 1, MessageType.MsgPropose, 1))
        r.persist()

        m = new_message(2, 1, MessageType.MsgAppendResponse, 0)
        m.set_term(r.raft.get_term())
        m.set_index(index)
        r.step(m)

        assert (
            r.raft_log.get_committed() == wcommit
        ), f"#{i}: commit = {r.raft_log.get_committed()}, want {wcommit}"



from typing import List

from rraft import (
    Message,
    MessageRef,
    InMemoryRaft,
    InMemoryRaftLogRef,
)


class Interface:
    """
    # A simulated Raft facade for testing.
    #
    # If the contained value is a `Some` operations happen. If they are a `None` operations are
    # a no-op.
    #
    # Compare to upstream, we use struct instead of trait here.
    # Because to be able to cast Interface later, we have to make
    # Raft derive Any, which will require a lot of dependencies to derive Any.
    # That's not worthy for just testing purpose.
    """

    def __init__(self, r: InMemoryRaft) -> None:
        """
        Create a new interface to a new raft.
        """
        self.raft = r

    def __repr__(self) -> str:
        if not self.raft:
            return "Interface {{ empty }}"
        return f"Interface {{ id: {self.raft.get_id()} }}"

    @property
    def raft_log(self) -> InMemoryRaftLogRef:
        return self.raft.get_raft_log()

    def step(self, message: Message | MessageRef) -> None:
        """
        Step the raft, if it exists.
        """
        if self.raft:
            self.raft.step(message)

    def read_messages(self) -> List[Message]:
        """
        Read messages out of the raft.
        """
        if self.raft:
            return self.raft.take_msgs()
        return []

    def persist(self) -> None:
        """
        Persist the unstable snapshot and entries.
        """
        if self.raft:
            if snapshot := self.raft_log.unstable_snapshot():
                snap = snapshot.clone()
                index = snap.get_metadata().get_index()
                self.raft_log.stable_snap(index)
                self.raft_log.get_store().wl().apply_snapshot(snap)
                self.raft.on_persist_snap(index)
                self.raft.commit_apply(index)

            if unstable := self.raft_log.unstable_entries():
                e = unstable[-1]
                cloned_unstable = list(map(lambda x: x.clone(), unstable))

                last_idx, last_term = e.get_index(), e.get_term()
                self.raft_log.stable_entries(last_idx, last_term)
                self.raft_log.get_store().wl().append(cloned_unstable)
                self.raft.on_persist_entries(last_idx, last_term)



import asyncio
from asyncio import Queue
from datetime import datetime, timezone
from typing import Callable, Dict, List
from rraft import (
    Config,
    ConfState,
    EntryRef,
    EntryType,
    LoggerRef,
    Logger,
    MemStorage,
    MessageRef,
    OverflowStrategy,
    InMemoryRawNode,
)

channel: Queue = Queue()


def now() -> int:
    return int(datetime.now(tz=timezone.utc).timestamp() * 1000)


async def send_propose(logger: Logger | LoggerRef) -> None:
    # Wait some time and send the request to the Raft.
    await asyncio.sleep(10)
    logger.info("propose a request")

    # Send a command to the Raft, wait for the Raft to apply it
    # and get the result.
    raft_chan: Queue = Queue()

    await channel.put(
        {
            "msg_type": "PROPOSE",
            "id": 1,
            "cb": lambda: raft_chan.put(0),
        }
    )

    n = await raft_chan.get()
    assert n == 0
    logger.info("receive the propose callback")

    await channel.put({"msg_type": "DISCONNECTED"})


async def on_ready(
    raft_group: InMemoryRawNode, cbs: Dict[str, Callable]
) -> None:
    if not raft_group.has_ready():
        return

    store = raft_group.get_raft().get_raft_log().get_store()

    # Get the `Ready` with `RawNode::ready` interface.
    ready = raft_group.ready()

    async def handle_messages(msgs: List[MessageRef]):
        for _msg in msgs:
            # Send messages to other peers.
            continue

    if msgs := ready.messages():
        # Send out the messages come from the node.
        await handle_messages(msgs)

    if ready.snapshot():
        # This is a snapshot, we need to apply the snapshot at first.
        cloned_ready = raft_group.ready()
        store.wl().apply_snapshot(cloned_ready.snapshot())

    _last_apply_index = 0

    async def handle_committed_entries(committed_entries: List[EntryRef]):
        for entry in committed_entries:
            # Mostly, you need to save the last apply index to resume applying
            # after restart. Here we just ignore this because we use a Memory storage.
            nonlocal _last_apply_index
            _last_apply_index = entry.get_index()

            entry_data = entry.get_data()

            if not entry.get_data():
                # Emtpy entry, when the peer becomes Leader it will send an empty entry.
                continue

            if entry.get_entry_type() == EntryType.EntryNormal:
                await cbs[entry_data[0]]()
                del cbs[entry_data[0]]

            # TODO: handle EntryConfChange

    await handle_committed_entries(ready.committed_entries())

    if entries := ready.entries():
        # Append entries to the Raft log.
        store.wl().append(entries)

    if hs := ready.hs():
        # Raft HardState changed, and we need to persist it.
        store.wl().set_hardstate(hs)

    if msgs := ready.persisted_messages():
        # Send out the persisted messages come from the node.
        await handle_messages(msgs)

    # Advance the Raft.
    light_rd = raft_group.advance(ready.make_ref())

    # Update commit index.
    if commit := light_rd.commit_index():
        store.wl().hard_state().set_commit(commit)

    # Send out the messages.
    await handle_messages(light_rd.messages())
    # Apply all committed entries.
    await handle_committed_entries(light_rd.committed_entries())
    # Advance the apply index.
    raft_group.advance_apply()


async def main():
    # Create a storage for Raft, and here we just use a simple memory storage.
    # You need to build your own persistent storage in your production.
    # Please check the Storage trait in src/storage.rs to see how to implement one.
    storage = MemStorage.new_with_conf_state(ConfState(voters=[1], learners=[]))

    # Create the configuration for the Raft node.
    cfg = Config(
        # The unique ID for the Raft node.
        id=1,
        # Election tick is for how long the follower may campaign again after
        # it doesn't receive any message from the leader.
        election_tick=10,
        # Heartbeat tick is for how long the leader needs to send
        # a heartbeat to keep alive.
        heartbeat_tick=3,
        # The max size limits the max size of each appended message. Mostly, 1 MB is enough.
        max_size_per_msg=1024 * 1024 * 1024,
        # Max inflight msgs that the leader sends messages to follower without
        # receiving ACKs.
        max_inflight_msgs=256,
        # The Raft applied index.
        # You need to save your applied index when you apply the committed Raft logs.
        applied=0,
    )

    logger = Logger(chan_size=4096, overflow_strategy=OverflowStrategy.Block)

    # Create the Raft node.
    raw_node = InMemoryRawNode(cfg, storage, logger)

    # Use another task to propose a Raft request.
    asyncio.create_task(send_propose(logger))

    t = now()
    timeout = 100
    # Use a dict to hold the `propose` callbacks.
    cbs: Dict[str, Callable] = {}

    # Loop forever to drive the Raft.
    while True:
        try:
            top = await asyncio.wait_for(channel.get(), timeout / 1000)
            msg_type = top["msg_type"]

            if msg_type == "PROPOSE":
                id, cb = top["id"], top["cb"]
                cbs[id] = cb
                raw_node.propose(context=[], data=[id])
            elif msg_type == "RAFT":
                # Here we don't use Raft Message, so there is no "msg" sender in this example.
                msg = top["msg"]
                raw_node.step(msg)
            elif msg_type == "DISCONNECTED":
                break

        except asyncio.exceptions.TimeoutError:
            pass

        finally:
            d = now() - t
            t = now()

            if d >= timeout:
                timeout = 100
                # We drive Raft every 100ms.
                raw_node.tick()
            else:
                timeout -= d

            await on_ready(raw_node, cbs)


if __name__ == "__main__":
    asyncio.run(main())
